from __future__ import annotations

import datetime
import logging
import time
from dataclasses import dataclass
from typing import List, Optional, Tuple, Dict, Sequence

import fuzzywuzzy.fuzz
import fuzzywuzzy.process

from metricflow.constraints.time_constraint import TimeRangeConstraint
from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.dataflow_plan import BaseOutput
from metricflow.dataset.data_source_adapter import DataSourceDataSet
from metricflow.dataset.dataset import DataSet
from metricflow.errors.errors import UnableToSatisfyQueryError
from metricflow.model.objects.constraints.where import WhereClauseConstraint
from metricflow.model.objects.elements.dimension import DimensionType
from metricflow.model.semantic_model import SemanticModel
from metricflow.model.spec_converters import WhereConstraintConverter
from metricflow.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow.object_utils import pformat_big_objects
from metricflow.query.query_exceptions import InvalidQueryException
from metricflow.references import DimensionReference, IdentifierReference, MetricReference, TimeDimensionReference
from metricflow.specs import (
    MetricFlowQuerySpec,
    MetricSpec,
    DimensionSpec,
    TimeDimensionSpec,
    IdentifierSpec,
    LinkableInstanceSpec,
    OrderBySpec,
    OutputColumnNameOverride,
    SpecWhereClauseConstraint,
    LinkableSpecSet,
)
from metricflow.time.time_granularity import TimeGranularity
from metricflow.time.time_granularity_solver import (
    TimeGranularitySolver,
    PartialTimeDimensionSpec,
    RequestTimeGranularityException,
)

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class LinkableInstanceSpecs:
    """All linkable specs."""

    dimension_specs: Tuple[DimensionSpec, ...]
    time_dimension_specs: Tuple[TimeDimensionSpec, ...]
    partial_time_dimension_specs: Tuple[PartialTimeDimensionSpec, ...]
    identifier_specs: Tuple[IdentifierSpec, ...]


class MetricFlowQueryParser:
    """Parse input strings from the user into a metric query specification.

    Definitions:
    element name - the name of an element (measure, dimension, identifier) in a data source, or a metric name.
    qualified name - an element name with prefixes and suffixes added to it that further describe transformations or
    conditions for the element to retrieve. e.g. "ds__month" is the "ds" time dimension at the "month" granularity. Or
    "user_id__country" is the "country" dimension that is retrieved by joining "user_id" to the measure data source.
    """

    # Prefix to use for indents when logging.
    _INDENT_PREFIX = "  "

    def __init__(  # noqa: D
        self,
        model: SemanticModel,
        source_nodes: Sequence[BaseOutput[DataSourceDataSet]],
        node_output_resolver: DataflowPlanNodeOutputDataSetResolver[DataSourceDataSet],
    ) -> None:
        self._model = model
        self._metric_semantics = model.metric_semantics
        self._data_source_semantics = model.data_source_semantics

        # Set up containers for known element names
        self._known_identifier_element_references = self._data_source_semantics.get_identifier_references()

        self._known_time_dimension_element_references = [DataSet.metric_time_dimension_reference()]
        self._known_dimension_element_references = []
        for dimension_reference in self._data_source_semantics.get_dimension_references():
            dimension = self._data_source_semantics.get_dimension(dimension_reference)
            if dimension.type == DimensionType.CATEGORICAL:
                self._known_dimension_element_references.append(dimension_reference)
            elif dimension.type == DimensionType.TIME:
                self._known_time_dimension_element_references.append(dimension_reference.time_dimension_reference)
            else:
                raise RuntimeError(f"Unhandled linkable type: {dimension.type}")

        self._known_metric_names = set(self._metric_semantics.metric_references)
        self._metric_time_dimension_reference = DataSet.metric_time_dimension_reference()
        self._time_granularity_solver = TimeGranularitySolver(
            semantic_model=self._model,
            source_nodes=source_nodes,
            node_output_resolver=node_output_resolver,
        )

    @staticmethod
    def _top_fuzzy_matches(
        item: str,
        candidate_items: Sequence[str],
        max_suggestions: int = 6,
        min_score: int = 50,
    ) -> Sequence[str]:
        """Return the top items (by edit distance) in candidate_items that fuzzy matches the given item."""
        # Rank choices by edit distance score.
        # extract() return a tuple like (name, score)
        top_ranked_suggestions = sorted(
            fuzzywuzzy.process.extract(
                # This scorer seems to return the best results.
                item,
                list(candidate_items),
                limit=max_suggestions,
                scorer=fuzzywuzzy.fuzz.token_set_ratio,
            ),
            # Put the highest scoring item at the top of the list.
            key=lambda x: x[1],
            reverse=True,
        )
        # If the scores are too low, then it looks nonsensical, so remove those.
        top_ranked_suggestions = [x for x in top_ranked_suggestions if x[1] > min_score]
        return [x[0] for x in top_ranked_suggestions]

    def parse_and_validate_query(
        self,
        metric_names: Sequence[str],
        group_by_names: Sequence[str],
        limit: Optional[int] = None,
        time_constraint_start: Optional[datetime.datetime] = None,
        time_constraint_end: Optional[datetime.datetime] = None,
        where_constraint: Optional[WhereClauseConstraint] = None,
        where_constraint_str: Optional[str] = None,
        order: Optional[Sequence[str]] = None,
        time_granularity: Optional[TimeGranularity] = None,
    ) -> MetricFlowQuerySpec:
        """Parse the query into spec objects, validating them in the process.

        e.g. make sure that the given metric is a valid metric name.
        """
        start_time = time.time()
        try:
            return self._parse_and_validate_query(
                metric_names=metric_names,
                group_by_names=group_by_names,
                limit=limit,
                time_constraint_start=time_constraint_start,
                time_constraint_end=time_constraint_end,
                where_constraint=where_constraint,
                where_constraint_str=where_constraint_str,
                order=order,
                time_granularity=time_granularity,
            )
        finally:
            logger.info(f"Parsing the query took: {time.time() - start_time:.2f}s")

    def _validate_linkable_specs(
        self,
        metric_references: Tuple[MetricReference, ...],
        all_linkable_specs: LinkableInstanceSpecs,
        time_dimension_specs: Tuple[TimeDimensionSpec, ...],
    ) -> None:
        invalid_group_bys = self._get_invalid_linkable_specs(
            metric_references=metric_references,
            dimension_specs=all_linkable_specs.dimension_specs,
            time_dimension_specs=time_dimension_specs,
            identifier_specs=all_linkable_specs.identifier_specs,
        )
        if len(invalid_group_bys) > 0:
            valid_group_by_names_for_metrics = sorted(
                x.qualified_name
                for x in self._metric_semantics.element_specs_for_metrics(metric_references=list(metric_references))
            )
            # Create suggestions for invalid dimensions in case the user made a typo.
            suggestion_sections = {}
            for invalid_group_by in invalid_group_bys:
                suggestions = MetricFlowQueryParser._top_fuzzy_matches(
                    item=invalid_group_by.qualified_name, candidate_items=valid_group_by_names_for_metrics
                )
                section_key = f"Suggestions for invalid dimension '{invalid_group_by.qualified_name}'"
                section_value = pformat_big_objects(suggestions)
                suggestion_sections[section_key] = section_value
            raise UnableToSatisfyQueryError(
                f"Dimensions {[x.qualified_name for x in invalid_group_bys]} cannot be "
                f"resolved for metrics {[x.element_name for x in metric_references]}. The invalid dimension may not "
                f"exist, require an ambiguous join (e.g. a join path that can be satisfied in multiple ways), "
                f"or require a fanout join.",
                context=suggestion_sections,
            )

    def _parse_and_validate_query(
        self,
        metric_names: Sequence[str],
        group_by_names: Sequence[str],
        limit: Optional[int] = None,
        time_constraint_start: Optional[datetime.datetime] = None,
        time_constraint_end: Optional[datetime.datetime] = None,
        where_constraint: Optional[WhereClauseConstraint] = None,
        where_constraint_str: Optional[str] = None,
        order: Optional[Sequence[str]] = None,
        time_granularity: Optional[TimeGranularity] = None,
    ) -> MetricFlowQuerySpec:
        assert not (
            where_constraint and where_constraint_str
        ), "Both where_constraint and where_constraint_str should not be set"

        if len(group_by_names) == 0:
            assert len(metric_names) == 1, "Querying for multiple metrics with no dimensions is not yet supported."

        parsed_where_constraint: Optional[WhereClauseConstraint]
        if where_constraint_str:
            parsed_where_constraint = WhereClauseConstraint.parse(where_constraint_str)
        else:
            parsed_where_constraint = where_constraint

        metric_references = self._parse_metric_names(metric_names)
        metric_specs = tuple(MetricSpec.from_reference(metric_reference) for metric_reference in metric_references)

        if time_constraint_start is None:
            time_constraint_start = TimeRangeConstraint.ALL_TIME_BEGIN()
        elif time_constraint_start < TimeRangeConstraint.ALL_TIME_BEGIN():
            logger.warning(
                f"Start time for the supplied time constraint {time_constraint_start.isoformat()} is <less> than the "
                f"minimum allowed ('{TimeRangeConstraint.ALL_TIME_BEGIN().isoformat()}'). Changing to the minimum."
            )
            time_constraint_start = TimeRangeConstraint.ALL_TIME_BEGIN()
        elif time_constraint_start > TimeRangeConstraint.ALL_TIME_END():
            logger.warning(
                f"Start time for the supplied time constraint {time_constraint_start.isoformat()} is > than the "
                f"maximum allowed ('{TimeRangeConstraint.ALL_TIME_END().isoformat()}'). Changing to the maximum."
            )
            time_constraint_start = TimeRangeConstraint.ALL_TIME_END()

        if time_constraint_end is None:
            time_constraint_end = TimeRangeConstraint.ALL_TIME_END()
        elif time_constraint_end > TimeRangeConstraint.ALL_TIME_END():
            logger.warning(
                f"End time for the supplied time constraint {time_constraint_end.isoformat()} is > than the "
                f"maximum allowed ('{TimeRangeConstraint.ALL_TIME_END().isoformat()}'). Changing to the maximum."
            )
            time_constraint_end = TimeRangeConstraint.ALL_TIME_END()
        elif time_constraint_end < TimeRangeConstraint.ALL_TIME_BEGIN():
            logger.warning(
                f"End time for the supplied time constraint {time_constraint_end.isoformat()} is less than the "
                f"minimum allowed ('{TimeRangeConstraint.ALL_TIME_BEGIN().isoformat()}'). Changing to the minimum."
            )
            time_constraint_end = TimeRangeConstraint.ALL_TIME_BEGIN()

        if time_constraint_end < time_constraint_start:
            raise ValueError(
                f"End of time constraint {time_constraint_end.isoformat()} is before start of time constraint "
                f"{time_constraint_start.isoformat()}"
            )

        time_constraint: Optional[TimeRangeConstraint] = TimeRangeConstraint(
            start_time=time_constraint_start, end_time=time_constraint_end
        )
        if time_constraint == TimeRangeConstraint.all_time():
            # If the time constraint is all time, just ignore and not render
            time_constraint = None

        requested_linkable_specs = self._parse_linkable_element_names(group_by_names, metric_references)
        partial_time_dimension_spec_replacements = (
            self._time_granularity_solver.resolve_granularity_for_partial_time_dimension_specs(
                metric_references=metric_references,
                partial_time_dimension_specs=requested_linkable_specs.partial_time_dimension_specs,
                metric_time_dimension_reference=self._metric_time_dimension_reference,
                time_granularity=time_granularity,
            )
        )

        all_group_by_names = list(group_by_names)
        if parsed_where_constraint is not None:
            all_group_by_names += parsed_where_constraint.linkable_names

        time_dimension_specs = requested_linkable_specs.time_dimension_specs + tuple(
            time_dimension_spec for _, time_dimension_spec in partial_time_dimension_spec_replacements.items()
        )

        self._time_granularity_solver.validate_time_granularity(metric_references, time_dimension_specs)

        order_by_specs = self._parse_order_by(order or [], partial_time_dimension_spec_replacements)

        for metric_reference in metric_references:
            metric = self._metric_semantics.get_metric(metric_reference)
            if metric.constraint is not None:
                all_linkable_specs = self._parse_linkable_element_names(
                    qualified_linkable_names=all_group_by_names + metric.constraint.linkable_names,
                    metric_references=(metric_reference,),
                )
                self._validate_linkable_specs(
                    metric_references=(metric_reference,),
                    all_linkable_specs=all_linkable_specs,
                    time_dimension_specs=time_dimension_specs,
                )
        all_linkable_specs = self._parse_linkable_element_names(
            qualified_linkable_names=all_group_by_names,
            metric_references=metric_references,
        )
        self._validate_linkable_specs(
            metric_references=metric_references,
            all_linkable_specs=all_linkable_specs,
            time_dimension_specs=time_dimension_specs,
        )

        self._validate_order_by_specs(
            order_by_specs=order_by_specs,
            metric_references=metric_references,
            linkable_specs=LinkableSpecSet(
                dimension_specs=requested_linkable_specs.dimension_specs,
                time_dimension_specs=time_dimension_specs,
                identifier_specs=requested_linkable_specs.identifier_specs,
            ),
        )

        # Update constraints to be appropriate for the time granularity of the query.
        if time_constraint:
            logger.info(f"Time constraint before adjustment is {time_constraint}")
            time_constraint = self._adjust_time_range_constraint(
                metric_references=metric_references,
                time_dimension_specs=time_dimension_specs,
                time_range_constraint=time_constraint,
            )
            logger.info(f"Time constraint after adjustment is {time_constraint}")

        # In some cases, the old framework does not use dundered suffixes in the output column name for the primary
        # time dimension. We should aim to get rid of this logic.
        output_column_name_overrides = []
        if (
            self._metric_time_dimension_specified_without_granularity(
                requested_linkable_specs.partial_time_dimension_specs
            )
            and not time_granularity
        ):
            if self._metrics_have_same_time_granularities(metric_references=metric_references):
                _, replace_with_time_dimension_spec = self._find_replacement_for_metric_time_dimension(
                    partial_time_dimension_spec_replacements
                )
                output_column_name_overrides.append(
                    OutputColumnNameOverride(
                        time_dimension_spec=replace_with_time_dimension_spec,
                        output_column_name=self._metric_time_dimension_reference.element_name,
                    )
                )

        if limit is not None and limit < 0:
            raise InvalidQueryException(f"Limit was specified as {limit}, which is < 0.")

        spec_where_constraint: Optional[SpecWhereClauseConstraint] = None
        if parsed_where_constraint:
            spec_where_constraint = WhereConstraintConverter.convert_to_spec_where_constraint(
                data_source_semantics=self._data_source_semantics,
                where_constraint=parsed_where_constraint,
            )

        return MetricFlowQuerySpec(
            metric_specs=metric_specs,
            dimension_specs=requested_linkable_specs.dimension_specs,
            identifier_specs=requested_linkable_specs.identifier_specs,
            time_dimension_specs=time_dimension_specs,
            order_by_specs=order_by_specs,
            output_column_name_overrides=tuple(output_column_name_overrides),
            time_range_constraint=time_constraint,
            where_constraint=spec_where_constraint,
            limit=limit,
        )

    def _validate_order_by_specs(
        self,
        order_by_specs: Sequence[OrderBySpec],
        metric_references: Sequence[MetricReference],
        linkable_specs: LinkableSpecSet,
    ) -> None:
        """Checks that the order by specs references an item in the query."""

        # TODO: this is a workaround
        # Need to figure out whether we should clean up OrderBySpec or if we have to actually pass a fully resolved MetricSpec here
        metric_specs = [MetricSpec.from_reference(metric_reference) for metric_reference in metric_references]
        for order_by_spec in order_by_specs:
            if not (
                order_by_spec.item in metric_specs
                or order_by_spec.item in linkable_specs.dimension_specs
                or order_by_spec.item in linkable_specs.time_dimension_specs
                or order_by_spec.item in linkable_specs.identifier_specs
            ):
                raise InvalidQueryException(f"Order by item {order_by_spec} not in the query")

    def _adjust_time_range_constraint(
        self,
        metric_references: Sequence[MetricReference],
        time_dimension_specs: Sequence[TimeDimensionSpec],
        time_range_constraint: TimeRangeConstraint,
    ) -> TimeRangeConstraint:
        """Adjust the time range constraint so that it matches the boundaries of the granularity of the result."""
        self._time_granularity_solver.validate_time_granularity(metric_references, time_dimension_specs)

        smallest_primary_time_granularity_in_query = self._find_smallest_metric_time_dimension_spec_granularity(
            time_dimension_specs
        )
        if smallest_primary_time_granularity_in_query:
            adjusted_to_granularity = smallest_primary_time_granularity_in_query

        else:
            _, adjusted_to_granularity = self._time_granularity_solver.local_dimension_granularity_range(
                metric_references=metric_references,
                local_time_dimension_reference=self._metric_time_dimension_reference,
            )
        logger.info(f"Adjusted primary time granularity is {adjusted_to_granularity}")
        return self._time_granularity_solver.adjust_time_range_to_granularity(
            time_range_constraint, adjusted_to_granularity
        )

    def _find_replacement_for_metric_time_dimension(
        self, replacements: Dict[PartialTimeDimensionSpec, TimeDimensionSpec]
    ) -> Tuple[PartialTimeDimensionSpec, TimeDimensionSpec]:
        for partial_time_dimension_spec_to_replace, replace_with_time_dimension_spec in replacements.items():
            if (
                partial_time_dimension_spec_to_replace.element_name
                == self._metric_time_dimension_reference.element_name
                and partial_time_dimension_spec_to_replace.identifier_links == ()
            ):
                return partial_time_dimension_spec_to_replace, replace_with_time_dimension_spec

        raise RuntimeError(f"Replacement for metric time dimension '{self._metric_time_dimension_reference}' not found")

    def _metric_time_dimension_specified_without_granularity(
        self, partial_time_dimension_specs: Sequence[PartialTimeDimensionSpec]
    ) -> bool:
        # This detects a user query like: "query --metrics monthly_bookings --dimensions ds"
        # The granularity for "ds" is not specified by the user.
        for time_dimension_spec in partial_time_dimension_specs:
            if (
                time_dimension_spec.element_name == self._metric_time_dimension_reference.element_name
                and time_dimension_spec.identifier_links == ()
            ):
                return True
        return False

    def _metrics_have_same_time_granularities(self, metric_references: Sequence[MetricReference]) -> bool:
        (min_granularity, max_granularity,) = self._time_granularity_solver.local_dimension_granularity_range(
            metric_references=metric_references,
            local_time_dimension_reference=self._metric_time_dimension_reference,
        )
        return min_granularity == max_granularity

    def _find_smallest_metric_time_dimension_spec_granularity(
        self, time_dimension_specs: Sequence[TimeDimensionSpec]
    ) -> Optional[TimeGranularity]:
        metric_time_dimension_specs: List[TimeDimensionSpec] = [
            x
            for x in time_dimension_specs
            if (
                x.element_name == self._metric_time_dimension_reference.element_name
                and x.identifier_links == ()
                and x.time_granularity
            )
        ]

        metric_time_dimension_specs.sort(key=lambda x: x.time_granularity.to_int())
        if len(metric_time_dimension_specs) > 0:
            return metric_time_dimension_specs[0].time_granularity
        else:
            return None

    def _parse_metric_names(self, metric_names: Sequence[str]) -> Tuple[MetricReference, ...]:
        """Converts metric names into metric names. An exception is thrown if the name is invalid."""

        # The config must be lower-case, so we lower case for case-insensitivity against query inputs from the user.
        metric_names = [x.lower() for x in metric_names]

        known_metric_names = set(self._metric_semantics.metric_references)
        metric_references: List[MetricReference] = []
        for metric_name in metric_names:
            metric_reference = MetricReference(element_name=metric_name)
            if metric_reference not in known_metric_names:
                suggestions = {
                    f"Suggestions for '{metric_name}'": pformat_big_objects(
                        MetricFlowQueryParser._top_fuzzy_matches(
                            item=metric_name,
                            candidate_items=[x.element_name for x in self._metric_semantics.metric_references],
                        )
                    )
                }
                raise UnableToSatisfyQueryError(
                    f"Unknown metric: '{metric_name}'",
                    context=suggestions,
                )
            metric_references.append(metric_reference)
        return tuple(metric_references)

    def _parse_linkable_element_names(
        self, qualified_linkable_names: Sequence[str], metric_references: Sequence[MetricReference]
    ) -> LinkableInstanceSpecs:
        """Convert the linkable spec names into the respective specification objects."""

        qualified_linkable_names = [x.lower() for x in qualified_linkable_names]

        dimension_specs = []
        time_dimension_specs = []
        partial_time_dimension_specs = []
        identifier_specs = []

        for qualified_name in qualified_linkable_names:
            structured_name = StructuredLinkableSpecName.from_name(qualified_name)
            element_name = structured_name.element_name
            identifier_links = tuple(IdentifierReference(element_name=x) for x in structured_name.identifier_link_names)
            # Create the spec based on the type of element referenced.
            if TimeDimensionReference(element_name=element_name) in self._known_time_dimension_element_references:
                if structured_name.time_granularity:
                    time_dimension_specs.append(
                        TimeDimensionSpec(
                            element_name=element_name,
                            identifier_links=identifier_links,
                            time_granularity=structured_name.time_granularity,
                        )
                    )
                else:
                    partial_time_dimension_specs.append(
                        PartialTimeDimensionSpec(
                            element_name=element_name,
                            identifier_links=identifier_links,
                        )
                    )
            elif DimensionReference(element_name=element_name) in self._known_dimension_element_references:
                dimension_specs.append(DimensionSpec(element_name=element_name, identifier_links=identifier_links))
            elif IdentifierReference(element_name=element_name) in self._known_identifier_element_references:
                identifier_specs.append(IdentifierSpec(element_name=element_name, identifier_links=identifier_links))
            else:
                valid_group_by_names_for_metrics = sorted(
                    x.qualified_name for x in self._metric_semantics.element_specs_for_metrics(list(metric_references))
                )

                suggestions = {
                    f"Suggestions for '{qualified_name}'": pformat_big_objects(
                        MetricFlowQueryParser._top_fuzzy_matches(
                            item=qualified_name,
                            candidate_items=valid_group_by_names_for_metrics,
                        )
                    )
                }
                raise UnableToSatisfyQueryError(
                    f"Unknown element name '{element_name}' in dimension name '{qualified_name}'",
                    context=suggestions,
                )

        return LinkableInstanceSpecs(
            dimension_specs=tuple(dimension_specs),
            time_dimension_specs=tuple(time_dimension_specs),
            partial_time_dimension_specs=tuple(partial_time_dimension_specs),
            identifier_specs=tuple(identifier_specs),
        )

    def _get_invalid_linkable_specs(
        self,
        metric_references: Tuple[MetricReference, ...],
        dimension_specs: Tuple[DimensionSpec, ...],
        time_dimension_specs: Tuple[TimeDimensionSpec, ...],
        identifier_specs: Tuple[IdentifierSpec, ...],
    ) -> List[LinkableInstanceSpec]:
        """Checks that each requested linkable instance can be retrieved for the given metric"""
        invalid_linkable_specs: List[LinkableInstanceSpec] = []
        # TODO: distinguish between dimensions that invalid via typo vs ambiguous join path
        valid_linkable_specs = self._metric_semantics.element_specs_for_metrics(
            metric_references=list(metric_references)
        )

        for dimension_spec in dimension_specs:
            if dimension_spec not in valid_linkable_specs:
                invalid_linkable_specs.append(dimension_spec)

        for identifier_spec in identifier_specs:
            if identifier_spec not in valid_linkable_specs:
                invalid_linkable_specs.append(identifier_spec)

        for time_dimension_spec in time_dimension_specs:
            if (
                time_dimension_spec not in valid_linkable_specs
                # Because the metric time dimension is a virtual dimension that's not in the model, it won't be included
                # in valid_linkable_specs.
                and time_dimension_spec.reference != DataSet.metric_time_dimension_reference()
            ):
                invalid_linkable_specs.append(time_dimension_spec)

        return invalid_linkable_specs

    def _parse_order_by(
        self,
        order_by_names: Sequence[str],
        time_dimension_spec_replacements: Dict[PartialTimeDimensionSpec, TimeDimensionSpec],
    ) -> Tuple[OrderBySpec, ...]:
        """time_dimension_spec_replacements is used to replace a partial spec from parsing the names to a full one."""
        # TODO: Validate identifier links
        # TODO: Validate order by items are in the query
        order_by_specs: List[OrderBySpec] = []
        for order_by_name in order_by_names:
            descending = False
            if order_by_name.startswith("-"):
                order_by_name = order_by_name[1:]
                descending = True
            parsed_name = StructuredLinkableSpecName.from_name(order_by_name)

            if MetricReference(element_name=parsed_name.element_name) in self._known_metric_names:
                if parsed_name.time_granularity:
                    raise InvalidQueryException(
                        f"Order by item '{order_by_name}' references a metric but has a time granularity"
                    )
                if parsed_name.identifier_link_names:
                    raise InvalidQueryException(
                        f"Order by item '{order_by_name}' references a metric but has identifier links"
                    )
                order_by_specs.append(
                    OrderBySpec(
                        metric_spec=MetricSpec(element_name=parsed_name.element_name),
                        descending=descending,
                    )
                )
            elif DimensionReference(element_name=parsed_name.element_name) in self._known_dimension_element_references:
                if parsed_name.time_granularity:
                    raise InvalidQueryException(
                        f"Order by item '{order_by_name}' references a categorical dimension but has a time "
                        f"granularity"
                    )
                order_by_specs.append(
                    OrderBySpec(
                        dimension_spec=DimensionSpec(
                            element_name=parsed_name.element_name,
                            identifier_links=tuple(
                                IdentifierReference(element_name=x) for x in parsed_name.identifier_link_names
                            ),
                        ),
                        descending=descending,
                    )
                )
            elif (
                TimeDimensionReference(element_name=parsed_name.element_name)
                in self._known_time_dimension_element_references
            ):
                identifier_links = tuple(IdentifierReference(element_name=x) for x in parsed_name.identifier_link_names)
                if parsed_name.time_granularity:
                    order_by_specs.append(
                        OrderBySpec(
                            time_dimension_spec=TimeDimensionSpec(
                                element_name=parsed_name.element_name,
                                identifier_links=identifier_links,
                                time_granularity=parsed_name.time_granularity,
                            ),
                            descending=descending,
                        )
                    )
                else:
                    # If the time granularity for an order by wasn't specified, replace it with the same time
                    # granularity as it was done for the requested dimensions.
                    partial_time_dimension_spec = PartialTimeDimensionSpec(
                        element_name=parsed_name.element_name,
                        identifier_links=identifier_links,
                    )

                    if partial_time_dimension_spec in time_dimension_spec_replacements:
                        order_by_specs.append(
                            OrderBySpec(
                                time_dimension_spec=time_dimension_spec_replacements[partial_time_dimension_spec],
                                descending=descending,
                            )
                        )
                    else:
                        raise RequestTimeGranularityException(
                            f"Order by item '{order_by_name}' does not specify a time granularity and it does not "
                            f"match a requested time dimension"
                        )

            elif (
                IdentifierReference(element_name=parsed_name.element_name) in self._known_identifier_element_references
            ):
                if parsed_name.time_granularity:
                    raise InvalidQueryException(
                        f"Order by item '{order_by_name}' references an identifier but has a time granularity"
                    )
                order_by_specs.append(
                    OrderBySpec(
                        identifier_spec=IdentifierSpec(
                            element_name=parsed_name.element_name,
                            identifier_links=tuple(
                                IdentifierReference(element_name=x) for x in parsed_name.identifier_link_names
                            ),
                        ),
                        descending=descending,
                    )
                )
            else:
                raise InvalidQueryException(f"Order by item '{order_by_name}' references an element that is not known")

        return tuple(order_by_specs)



"""Protocols for classes meant to manage semantic information of useful object types.

These are useful as more generic descriptors of interfaces around classes like semantic containers and such, which
might be best used in a read-only mode. These containers could also pull in extra dependencies in their internal
implementations, dependencies which are best isolated inside the packages containing the concrete objects rather
than the interface specifications.
"""

from __future__ import annotations

from abc import abstractmethod
from typing import Dict, FrozenSet, List, Optional, Protocol, Sequence, Set, Tuple

from metricflow.instances import DataSourceElementReference, DataSourceReference
from metricflow.model.objects.data_source import DataSource, DataSourceOrigin
from metricflow.model.objects.elements.dimension import Dimension
from metricflow.model.objects.elements.identifier import Identifier
from metricflow.model.objects.elements.measure import Measure
from metricflow.model.objects.metric import Metric
from metricflow.model.semantics.element_group import ElementGrouper
from metricflow.model.semantics.linkable_spec_resolver import LinkableElementProperties
from metricflow.references import (
    DimensionReference,
    IdentifierReference,
    MeasureReference,
    MetricReference,
    TimeDimensionReference,
)
from metricflow.specs import (
    LinkableInstanceSpec,
    MeasureSpec,
    MetricInputMeasureSpec,
    NonAdditiveDimensionSpec,
)


class DataSourceSemanticsAccessor(Protocol):
    """Protocol defining core interface for accessing semantic information about a set of data source objects

    This is primarily useful for restricting caller access to the subset of container methods and imports we want
    them to use. For example, the DataSourceSemantics class might implement this protocol but also include some
    public methods for adding or removing data sources from the container, while this protocol only allows the
    caller to invoke the accessor methods which retrieve semantic information about the collected data sources.
    """

    @abstractmethod
    def get_dimension_references(self) -> List[DimensionReference]:
        """Retrieve all dimension references from the collection of data sources"""
        raise NotImplementedError

    @abstractmethod
    def get_dimension(
        self, dimension_reference: DimensionReference, origin: Optional[DataSourceOrigin] = None
    ) -> Dimension:
        """Retrieve the dimension model object associated with the time dimension reference"""
        raise NotImplementedError

    @abstractmethod
    def get_time_dimension(self, time_dimension_reference: TimeDimensionReference) -> Dimension:
        """Retrieve the dimension model object associated with the time dimension reference"""
        raise NotImplementedError

    @property
    @abstractmethod
    def measure_references(self) -> List[MeasureReference]:
        """Return all measure references from the collection of data sources"""
        raise NotImplementedError

    @property
    @abstractmethod
    def non_additive_dimension_specs_by_measure(self) -> Dict[MeasureReference, NonAdditiveDimensionSpec]:
        """Return a mapping from all semi-additive measures to their corresponding non additive dimension parameters

        This includes all measures with non-additive dimension parameters, if any, from the collection of data sources.
        """
        raise NotImplementedError

    @abstractmethod
    def get_measure(self, measure_reference: MeasureReference) -> Measure:
        """Retrieve the measure model object associated with the measure reference"""
        raise NotImplementedError

    @abstractmethod
    def get_identifier_references(self) -> List[IdentifierReference]:
        """Retrieve all identifier references from the collection of data sources"""
        raise NotImplementedError

    @abstractmethod
    def get_data_sources_for_measure(self, measure_reference: MeasureReference) -> List[DataSource]:
        """Retrieve a list of all data source model objects associated with the measure reference"""
        raise NotImplementedError

    @abstractmethod
    def get_agg_time_dimension_for_measure(self, measure_reference: MeasureReference) -> TimeDimensionReference:
        """Retrieves the aggregate time dimension that is associated with the measure reference"""

    @abstractmethod
    def get_identifier_in_data_source(self, ref: DataSourceElementReference) -> Optional[Identifier]:
        """Retrieve the identifier matching the element -> data source mapping, if any"""
        raise NotImplementedError

    @abstractmethod
    def get(self, data_source_name: str) -> Optional[DataSource]:
        """Retrieve the data source model object matching the given name, if any"""
        raise NotImplementedError

    @abstractmethod
    def get_by_reference(self, data_source_reference: DataSourceReference) -> Optional[DataSource]:
        """Retrieve the data source model object matching the input data source reference, if any"""
        raise NotImplementedError

    @property
    @abstractmethod
    def data_source_references(self) -> Sequence[DataSourceReference]:
        """Return all DataSourceReference objects associated with the data sources in the collection"""
        raise NotImplementedError

    @abstractmethod
    def get_aggregation_time_dimensions_with_measures(
        self, data_source_reference: DataSourceReference
    ) -> ElementGrouper[TimeDimensionReference, MeasureSpec]:
        """Return all aggregation time dimensions in the given data source with their associated measures"""
        raise NotImplementedError

    @abstractmethod
    def get_data_sources_for_identifier(self, identifier_reference: IdentifierReference) -> Set[DataSource]:
        """Return all data sources associated with an identifier reference"""
        raise NotImplementedError


class MetricSemanticsAccessor(Protocol):
    """Protocol defining core interface for accessing semantic information about a set of metric objects

    This is primarily useful for restricting caller access to the subset of container methods and imports we want
    them to use. For example, the MetricSemantics class might implement this protocol but also include some
    public methods for adding or removing metrics from the container, while this protocol only allows the
    caller to invoke the accessor methods which retrieve semantic information about the collected metrics.
    """

    @abstractmethod
    def element_specs_for_metrics(
        self,
        metric_references: List[MetricReference],
        with_any_property: FrozenSet[LinkableElementProperties] = LinkableElementProperties.all_properties(),
        without_any_property: FrozenSet[LinkableElementProperties] = frozenset(),
    ) -> List[LinkableInstanceSpec]:
        """Retrieve the matching set of linkable elements common to all metrics requested (intersection)"""
        raise NotImplementedError

    @abstractmethod
    def get_metrics(self, metric_references: List[MetricReference]) -> List[Metric]:
        """Retrieve the Metric model objects associated with the provided metric specs"""
        raise NotImplementedError

    @property
    @abstractmethod
    def metric_references(self) -> List[MetricReference]:
        """Return the metric references"""
        raise NotImplementedError

    @abstractmethod
    def get_metric(self, metric_reference: MetricReference) -> Metric:  # noqa:D
        raise NotImplementedError

    @property
    @abstractmethod
    def valid_hashes(self) -> Set[str]:
        """Return all of the hashes of the metric definitions."""
        raise NotImplementedError

    @abstractmethod
    def measures_for_metric(self, metric_reference: MetricReference) -> Tuple[MetricInputMeasureSpec, ...]:
        """Return the measure specs required to compute the metric."""
        raise NotImplementedError

    @abstractmethod
    def contains_cumulative_metric(self, metric_references: Sequence[MetricReference]) -> bool:
        """Returns true if any of the specs correspond to a cumulative metric."""
        raise NotImplementedError



from metricflow.model.objects.constraints.where import WhereClauseConstraint


def test_where_constraint_parsing() -> None:
    """Simple test of a parsed where with a function and some boolean and equality checks"""
    parsed_where = WhereClauseConstraint.parse("ds >= CAST('2020-01-01' AS TIMESTAMP) AND is_instant")
    assert set(parsed_where.linkable_names) == {"ds", "is_instant"}


def test_where_constraint_parsing_empty_function() -> None:
    """Test involving an empty function, which produces an empty object"""
    parsed_where = WhereClauseConstraint.parse("is_internal IS FALSE AND is_instant = false AND ds = CURRENT_DATE()")
    assert set(parsed_where.linkable_names) == {"is_internal", "is_instant", "ds"}


def test_where_constraint_with_between() -> None:
    """Testing a where constraint with a BETWEEN expression"""
    parsed_where = WhereClauseConstraint.parse("WHERE ds < CURRENT_DATE() AND price BETWEEN min_price AND 1.50")
    assert set(parsed_where.linkable_names) == {"ds", "price", "min_price"}



import pytest

from metricflow.api.metricflow_client import MetricFlowClient
from metricflow.model.semantic_model import SemanticModel
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.protocols.sql_client import SqlClient
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState


@pytest.fixture
def mf_client(
    create_simple_model_tables: bool,
    sql_client: SqlClient,
    simple_semantic_model: SemanticModel,
    time_spine_source: TimeSpineSource,
    mf_test_session_state: MetricFlowTestSessionState,
) -> MetricFlowClient:
    """Fixture for MetricFlowClient."""
    return MetricFlowClient(
        sql_client=sql_client,
        user_configured_model=simple_semantic_model.user_configured_model,
        system_schema=mf_test_session_state.mf_system_schema,
    )



from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import ClassVar, Dict, Optional, Protocol, Sequence

from pandas import DataFrame

from metricflow.dataflow.sql_table import SqlTable
from metricflow.sql.render.sql_plan_renderer import SqlQueryPlanRenderer
from metricflow.sql.sql_bind_parameters import SqlBindParameters


class SqlEngine(Enum):
    """Enumeration of SQL engines, including ones that are not yet supported."""

    BIGQUERY = "BigQuery"
    DUCKDB = "DuckDB"
    REDSHIFT = "Redshift"
    POSTGRES = "Postgres"
    SNOWFLAKE = "Snowflake"
    DATABRICKS = "Databricks"

    # Not yet supported.
    MYSQL = "MySQL"


class SqlClient(Protocol):
    """Base interface for SqlClient instances used inside MetricFlow.

    This provides the methods needed to execute SQL queries against the relevant Data Warehouse configuration.
    """

    @property
    @abstractmethod
    def sql_engine_attributes(self) -> SqlEngineAttributes:
        """Return a struct of engine-specific metadata.

        See documentation on SqlEngineAttributes for details. This property is configured as a tagged
        method rather than a simple property because defining it as a simple property makes it settable,
        and we should not be re-using SqlClient instances with different SqlEngineAttributes.
        """
        raise NotImplementedError

    @abstractmethod
    def create_table_as_select(
        self,
        sql_table: SqlTable,
        select_query: str,
        sql_bind_parameters: SqlBindParameters = SqlBindParameters(),
    ) -> None:
        """Method for creating a table from the provided select query

        Args:
            sql_table: The SqlTable metadata of the table to create
            select_query: The query to use to populate the table
            sql_bind_parameters: Map of values to substitute in to parameterized sql query strings
        """
        raise NotImplementedError

    @abstractmethod
    def create_table_from_dataframe(
        self,
        sql_table: SqlTable,
        df: DataFrame,
        chunk_size: Optional[int] = None,
    ) -> None:
        """Creates a table and populates it with the contents of the dataframe

        Args:
            sql_table: The SqlTable metadata of the table to create
            df: The Pandas DataFrame with the contents of the target table
            chunk_size: The number of rows to write per query
        """
        raise NotImplementedError

    @abstractmethod
    def query(
        self,
        stmt: str,
        sql_bind_parameters: SqlBindParameters = SqlBindParameters(),
    ) -> DataFrame:
        """Base query method, upon execution will run a query that returns a pandas DataFrame"""
        raise NotImplementedError

    @abstractmethod
    def execute(
        self,
        stmt: str,
        sql_bind_parameters: SqlBindParameters = SqlBindParameters(),
    ) -> None:
        """Base execute method."""
        raise NotImplementedError

    @abstractmethod
    def dry_run(
        self,
        stmt: str,
        sql_bind_parameters: SqlBindParameters = SqlBindParameters(),
    ) -> None:
        """Base dry_run method"""
        raise NotImplementedError

    @abstractmethod
    def list_tables(self, schema_name: str) -> Sequence[str]:
        """List the tables in the given schema"""
        raise NotImplementedError

    @abstractmethod
    def table_exists(self, sql_table: SqlTable) -> bool:
        """Determines whether or not the given table exists in the data warehouse"""
        raise NotImplementedError

    @abstractmethod
    def drop_table(self, sql_table: SqlTable) -> None:
        """Drop the given table from the data warehouse"""
        raise NotImplementedError

    @abstractmethod
    def create_schema(self, schema_name: str) -> None:
        """Create the given schema if it doesn't already exist"""
        raise NotImplementedError

    @abstractmethod
    def drop_schema(self, schema_name: str, cascade: bool) -> None:  # noqa: D
        """Drop the given schema if it exists. If cascade is set, drop the tables in the schema first."""
        raise NotImplementedError

    @abstractmethod
    def health_checks(self, schema_name: str) -> Dict[str, Dict[str, str]]:
        """Run health checks against the underlying Data Warehouse

        TODO: Consider restructuring this so the health checks are separate from the SqlClient
        TODO: Re-evaluate the return type to see if there's a more structured option available

        This method fires a bunch of queries and collects results inside of a flexible output object for
        later processing. This method has been ported over to the SqlClient Protocol from the original
        closed-source client classes. It was included there because certain health checks are warehouse-specific,
        e.g., there are checks to verify that specific session values are properly configured for Snowflake
        instances, and migrating to a new model will be quite messy in the short term. Longer term it likely
        makes sense for these to be in a separate module, but for the time being the CLI will invoke health
        checks and therefore this needs to remain part of the core Protocol.
        """
        raise NotImplementedError

    @abstractmethod
    def close(self) -> None:  # noqa: D
        """Close the connections / engines used by this client."""
        raise NotImplementedError

    @abstractmethod
    def cancel_submitted_queries(self) -> None:  # noqa: D
        """Cancel queries submitted through this client (that may be still running) with best-effort."""
        raise NotImplementedError

    @abstractmethod
    def render_execution_param_key(self, execution_param_key: str) -> str:
        """Wrap execution parameter key with syntax accepted by engine."""
        raise NotImplementedError


class SqlEngineAttributes(Protocol):
    """Base interface for SQL engine-specific attributes and features

    These include items like support for language features (e.g., FULL OUTER JOIN support), dialect differences
    (e.g., DOUBLE type name), and things of that nature.

    Concrete implementations would typically be the equivalent of frozen dataclass literals, one per
    MetricFlowSupportedDBEngine, as we would not expect these properties to change from one client to the next.

    Concrete implementations SHOULD NOT inherit from this protocol, as the typechecker may not catch issues
    caused by changes to the protocol itself when inheritance is used.
    """

    sql_engine_type: ClassVar[SqlEngine]

    # SQL Engine capabilities
    date_trunc_supported: ClassVar[bool]
    full_outer_joins_supported: ClassVar[bool]
    indexes_supported: ClassVar[bool]
    multi_threading_supported: ClassVar[bool]
    timestamp_type_supported: ClassVar[bool]
    timestamp_to_string_comparison_supported: ClassVar[bool]
    cancel_submitted_queries_supported: ClassVar[bool]

    # SQL Dialect replacement strings
    double_data_type_name: ClassVar[str]
    timestamp_type_name: ClassVar[Optional[str]]

    # MetricFlow attributes
    sql_query_plan_renderer: ClassVar[SqlQueryPlanRenderer]



from datetime import datetime
from unittest.mock import patch, MagicMock

from metricflow.cli.cli_context import CLIContext
from metricflow.cli.main import (
    drop_materialization,
    get_dimension_values,
    health_checks,
    list_dimensions,
    list_materializations,
    list_metrics,
    materialize,
    query,
    validate_configs,
    version,
)
from metricflow.model.model_validator import ModelValidator
from metricflow.model.parsing.config_linter import ConfigLinter
from metricflow.model.validations.validator_helpers import (
    ModelValidationResults,
    ValidationError,
    ValidationFutureError,
    ValidationWarning,
)
from metricflow.test.fixtures.cli_fixtures import MetricFlowCliRunner


def test_query(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(query, args=["--metrics", "bookings", "--dimensions", "ds"])
    assert "bookings" in resp.output
    assert resp.exit_code == 0


def test_list_dimensions(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(list_dimensions, args=["--metric-names", "bookings"])

    assert "ds" in resp.output
    assert resp.exit_code == 0


def test_list_metrics(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(list_metrics)

    assert "bookings_per_listing: ds" in resp.output
    assert resp.exit_code == 0


def test_get_dimension_values(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(get_dimension_values, args=["--metric-name", "bookings", "--dimension-name", "is_instant"])

    actual_output_lines = sorted(resp.output.split("\n"))
    assert ["", "• False", "• True"] == actual_output_lines
    assert resp.exit_code == 0


def test_list_materializations(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(list_materializations)

    assert "materialization: details related to materialization" in resp.output
    assert resp.exit_code == 0


def test_materialize(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(
        materialize, args=["--materialization-name", "test_materialization", "--start-time", "2020-01-02"]
    )

    assert "Materialized table created at" in resp.output
    assert "test_materialization" in resp.output
    assert resp.exit_code == 0


def test_drop_materialization(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(drop_materialization, args=["--materialization-name", "test_materialization"])

    assert resp.exit_code == 0


def test_validate_configs(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    # Mock build result for `model_build_result_from_config`
    mocked_parsing_result = MagicMock(issues=ModelValidationResults())
    # Mock validation errors in validate_model function
    issues = (
        ValidationWarning(context=None, message="warning_message"),  # type: ignore
        ValidationFutureError(context=None, message="future_error_message", error_date=datetime.now()),  # type: ignore
        ValidationError(context=None, message="error_message"),  # type: ignore
    )
    mocked_build_result = MagicMock(issues=ModelValidationResults.from_issues_sequence(issues))
    with patch("metricflow.cli.main.model_build_result_from_config", return_value=mocked_parsing_result):
        with patch("metricflow.cli.main.path_to_models", return_value=""):
            with patch.object(ModelValidator, "validate_model", return_value=mocked_build_result):
                resp = cli_runner.run(validate_configs)

    assert "error_message" in resp.output
    assert resp.exit_code == 0


def test_future_errors_and_warnings_conditionally_show_up(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    # Mock build result for `model_build_result_from_config`
    mocked_parsing_result = MagicMock(issues=ModelValidationResults())
    # Mock validation errors in validate_model function
    issues = (
        ValidationWarning(context=None, message="warning_message"),  # type: ignore
        ValidationFutureError(context=None, message="future_error_message", error_date=datetime.now()),  # type: ignore
        ValidationError(context=None, message="error_message"),  # type: ignore
    )
    mocked_build_result = MagicMock(issues=ModelValidationResults.from_issues_sequence(issues))
    with patch("metricflow.cli.main.model_build_result_from_config", return_value=mocked_parsing_result):
        with patch("metricflow.cli.main.path_to_models", return_value=""):
            with patch.object(ModelValidator, "validate_model", return_value=mocked_build_result):
                resp = cli_runner.run(validate_configs)

    assert "warning_message" not in resp.output
    assert "future_error_message" not in resp.output
    assert resp.exit_code == 0

    with patch("metricflow.cli.main.model_build_result_from_config", return_value=mocked_parsing_result):
        with patch("metricflow.cli.main.path_to_models", return_value=""):
            with patch.object(ModelValidator, "validate_model", return_value=mocked_build_result):
                resp = cli_runner.run(validate_configs, ["--show-all"])

    assert "warning_message" in resp.output
    assert "future_error_message" in resp.output
    assert resp.exit_code == 0


def test_validate_configs_data_warehouse_validations(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    # Mock build result for `model_build_result_from_config`
    mocked_parsing_result = MagicMock(issues=ModelValidationResults())
    dw_validation_issues = [
        ValidationError(context=None, message="Data Warehouse Error"),  # type: ignore
    ]
    with patch("metricflow.cli.main.model_build_result_from_config", return_value=mocked_parsing_result):
        with patch("metricflow.cli.main.path_to_models", return_value=""):
            with patch.object(
                ModelValidator, "validate_model", return_value=MagicMock(issues=ModelValidationResults())
            ):
                with patch.object(CLIContext, "sql_client", return_value=None):  # type: ignore
                    with patch(
                        "metricflow.cli.main._run_dw_validations",
                        return_value=ModelValidationResults(errors=dw_validation_issues),
                    ):
                        resp = cli_runner.run(validate_configs)

    assert "Data Warehouse Error" in resp.output
    assert resp.exit_code == 0


def test_validate_configs_skip_data_warehouse_validations(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    # Mock build result for `model_build_result_from_config`
    mocked_parsing_result = MagicMock(issues=ModelValidationResults())
    with patch("metricflow.cli.main.model_build_result_from_config", return_value=mocked_parsing_result):
        with patch("metricflow.cli.main.path_to_models", return_value=""):
            with patch.object(
                ModelValidator, "validate_model", return_value=MagicMock(issues=ModelValidationResults())
            ):
                resp = cli_runner.run(validate_configs, args=["--skip-dw"])

    assert "Data Warehouse Error" not in resp.output
    assert resp.exit_code == 0


def test_validate_configs_with_lint_issues(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    lint_issues = [ValidationError(context=None, message="YAML Lint Error")]
    with patch("metricflow.cli.main.path_to_models", return_value=""):
        with patch.object(ConfigLinter, "lint_dir", return_value=ModelValidationResults(errors=lint_issues)):
            resp = cli_runner.run(validate_configs)

    assert "YAML Lint Error" in resp.output
    assert resp.exit_code == 0


def test_version(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(version)

    assert resp.output
    assert resp.exit_code == 0


def test_health_checks(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(health_checks)

    assert "SELECT 1: Success!" in resp.output
    assert resp.exit_code == 0



"""Prefixes to use when generating IDs. Listed here to avoid conflicts."""

# Nodes in a DAG have IDs of the form "prefix_number". These are the prefixes used for the different node types.
import logging
import threading

from typing import Dict, Type

DATAFLOW_NODE_AGGREGATE_MEASURES_ID_PREFIX = "am"
DATAFLOW_NODE_COMPUTE_METRICS_ID_PREFIX = "cm"
DATAFLOW_NODE_JOIN_AGGREGATED_MEASURES_BY_GROUPBY_COLUMNS_PREFIX = "jamgc"
DATAFLOW_NODE_JOIN_TO_STANDARD_OUTPUT_ID_PREFIX = "jso"
DATAFLOW_NODE_JOIN_SELF_OVER_TIME_RANGE_ID_PREFIX = "jotr"
DATAFLOW_NODE_ORDER_BY_LIMIT_ID_PREFIX = "obl"
DATAFLOW_NODE_PASS_FILTER_ELEMENTS_ID_PREFIX = "pfe"
DATAFLOW_NODE_READ_SQL_SOURCE_ID_PREFIX = "rss"
DATAFLOW_NODE_WHERE_CONSTRAINT_ID_PREFIX = "wcc"
DATAFLOW_NODE_WRITE_TO_RESULT_DATAFRAME_ID_PREFIX = "wrd"
DATAFLOW_NODE_WRITE_TO_RESULT_TABLE_ID_PREFIX = "wrt"
DATAFLOW_NODE_COMBINE_METRICS_ID_PREFIX = "cbm"
DATAFLOW_NODE_CONSTRAIN_TIME_RANGE_ID_PREFIX = "ctr"
DATAFLOW_NODE_SET_MEASURE_AGGREGATION_TIME = "sma"
DATAFLOW_NODE_SEMI_ADDITIVE_JOIN_ID_PREFIX = "saj"

SQL_EXPR_COLUMN_REFERENCE_ID_PREFIX = "cr"
SQL_EXPR_COMPARISON_ID_PREFIX = "cmp"
SQL_EXPR_FUNCTION_ID_PREFIX = "fnc"
SQL_EXPR_STRING_ID_PREFIX = "str"
SQL_EXPR_NULL_PREFIX = "null"
SQL_EXPR_LOGICAL_OPERATOR_PREFIX = "lo"
SQL_EXPR_STRING_LITERAL_PREFIX = "sl"
SQL_EXPR_IS_NULL_PREFIX = "isn"
SQL_EXPR_CAST_TO_TIMESTAMP_PREFIX = "ctt"
SQL_EXPR_DATE_TRUNC = "dt"
SQL_EXPR_RATIO_COMPUTATION = "rc"
SQL_EXPR_BETWEEN_PREFIX = "betw"

SQL_PLAN_SELECT_STATEMENT_ID_PREFIX = "ss"
SQL_PLAN_TABLE_FROM_CLAUSE_ID_PREFIX = "tfc"

EXEC_NODE_READ_SQL_QUERY = "rsq"
EXEC_NODE_NOOP = "noop"
EXEC_NODE_WRITE_TO_TABLE = "wtt"

DATAFLOW_PLAN_PREFIX = "dfp"
SQL_QUERY_PLAN_PREFIX = "sqp"
EXEC_PLAN_PREFIX = "ep"


logger = logging.getLogger(__name__)


class IdGenerator:
    """Helps generate unique ID strings."""

    def __init__(self, start_value: int = 0) -> None:  # noqa: D
        """Identifiers of the form (prefix + number) e.g. my_node_1"""
        self._next_id = start_value
        self._lock = threading.Lock()

    def create_id(self, prefix: str) -> str:
        """Create a unique ID string that can be used for naming nodes (or others) using the form <prefix>_<number>"""
        with self._lock:
            new_id = f"{prefix}_{self._next_id}"
            self._next_id += 1
            return new_id


class IdGeneratorRegistry:
    """Enumerate all IdGenerators used so that they can be patched appropriately in testing.

    See: patch_id_generators()
    """

    DEFAULT_START_VALUE = 0
    _state_lock = threading.Lock()
    _class_name_to_id_generator: Dict[str, IdGenerator] = {}

    @classmethod
    def for_class(cls, class_type: Type) -> IdGenerator:
        """Return an ID generator for the given class"""

        with cls._state_lock:
            class_name = ".".join([class_type.__module__, class_type.__name__])
            if class_name not in cls._class_name_to_id_generator:
                cls._class_name_to_id_generator[class_name] = IdGenerator(
                    start_value=IdGeneratorRegistry.DEFAULT_START_VALUE
                )
            return cls._class_name_to_id_generator[class_name]



"""Module with common models that are used across multiple classes/modules in inference."""

from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Optional, List
from enum import Enum

from metricflow.dataflow.sql_column import SqlColumn


class InferenceSignalConfidence(Enum):
    """A discrete enumeration of possible confidence values for an inference signal.

    We chose discrete confidence values instead of a continuous range (e.g a float between 0 and 1)
    to standardize confidence outputs between heuristic rules. We want to avoid different rules
    assuming different values for, say, "medium" or "high" confidence, since this could skew
    results in favor of/against certain rules.
    """

    VERY_HIGH = 3
    HIGH = 2
    MEDIUM = 1
    LOW = 0


class InferenceSignalNode(ABC):
    """A node in the inference signal type hierarchy.

    This class can be used to assemble a type hierarchy tree. It can be used by heuristics
    to check whether signals produced by rules are conflicting or complementary, relying
    on the property that sibling nodes are mutually exclusive in the hierarchy.
    """

    def __init__(self, parent: Optional[InferenceSignalNode], name: str) -> None:  # noqa: D
        self.name = name

        self.parent = parent
        self.children: List[InferenceSignalNode] = []

        if parent is not None:
            parent.children.append(self)

    def __str__(self) -> str:  # noqa: D
        return f"InferenceSignalNode: {self.name}"

    @property
    def supertypes(self) -> List[InferenceSignalNode]:
        """The list of all supertypes for this type node, from the root to the direct parent."""
        if self.parent is None:
            return []

        return self.parent.supertypes + [self.parent]

    def is_subtype_of(self, other: InferenceSignalNode) -> bool:
        """Whether self is a subtype of other.

        A subtype can always be used where its supertype is expected, although the reverse
        may not be true
        """
        return other == self or other in self.supertypes


# This is kinda horrible but there's no way of instancing the tree with type safety without
# hardcoding it. Having some magic that dynamically assigns attributes could work, but then
# we lose IDE autocompletion and static checking
class _TreeNodes:  # noqa: D
    root = InferenceSignalNode(None, "UNKNOWN")
    id = InferenceSignalNode(root, "IDENTIFIER")
    foreign_id = InferenceSignalNode(id, "FOREIGN_IDENTIFIER")
    unique_id = InferenceSignalNode(id, "UNIQUE_IDENTIFIER")
    primary_id = InferenceSignalNode(unique_id, "PRIMARY_IDENTIFIER")
    dimension = InferenceSignalNode(root, "DIMENSION")
    time_dimension = InferenceSignalNode(dimension, "TIME_DIMENSION")
    primary_time_dimension = InferenceSignalNode(time_dimension, "PRIMARY_TIME_DIMENSION")
    categorical_dimension = InferenceSignalNode(dimension, "CATEGORICAL_DIMENSION")
    measure = InferenceSignalNode(root, "MEASURE")


class InferenceSignalType:
    """All possible inference signal types."""

    UNKNOWN = _TreeNodes.root

    class ID:
        """Indicates a column was inferred to be an ID."""

        UNKNOWN = _TreeNodes.id
        FOREIGN = _TreeNodes.foreign_id
        UNIQUE = _TreeNodes.unique_id
        PRIMARY = _TreeNodes.primary_id

    class DIMENSION:
        """Indicates a column was inferred to be a dimension."""

        UNKNOWN = _TreeNodes.dimension
        TIME = _TreeNodes.time_dimension
        PRIMARY_TIME = _TreeNodes.primary_time_dimension
        CATEGORICAL = _TreeNodes.categorical_dimension

    class MEASURE:
        """Indicates a column was inferred to be a measure."""

        UNKNOWN = _TreeNodes.measure


@dataclass(frozen=True)
class InferenceSignal:
    """Encapsulates a piece of evidence about a column produced by an inference rule.

    column: the target column for this signal
    type_node: the type node of the signal
    reason: a human-readable string that explains why this signal was produced. It may
        eventually reach the user's eyeballs.
    confidence: the confidence that this signal is correct.
    only_applies_to_parent: whether a solver should only consider this signal if the parent type
        node is also present. In other words, this signal is _complementary_ to its parent.

    About the usage of `only_applies_to_parent`:
        This can be used to produce signals which don't affect each other, are not
        individually indicative of a specific outcome, and as such are only used to
        further specify parent signals.

        Example: columns with unique values can indicate both unique keys and categorical
        dimensions, which are contradictory outcomes. As such, a unique values signal requires
        the parent signal (is ID or is DIMENSION) to be present to be useful, and so anything
        tagged with this flag will only apply to the parent. So you can safely designate a
        signal as having a HIGH confidence for a given base column type while ignoring
        it for any other column type.


    """

    column: SqlColumn
    type_node: InferenceSignalNode
    reason: str
    confidence: InferenceSignalConfidence
    only_applies_to_parent: bool


@dataclass(frozen=True)
class InferenceResult:
    """Encapsulates a final decision about a column.

    column: the target column for this result
    type_node: the type node of the column
    reason: a list of human-readable strings that explain positive reasons why this result
        was produced. They may eventually reach the user's eyeballs.
    problems: a list of human-readable strings that explain why a more specific
        result was not produced or any other confusion/errors it might have encountered
        while solving. They may eventually reach the user's eyeballs.
    """

    column: SqlColumn
    type_node: InferenceSignalNode
    reasons: List[str]
    problems: List[str]



import logging
from typing import Optional, Tuple

from metricflow.aggregation_properties import AggregationState
from metricflow.column_assoc import (
    SingleColumnCorrelationKey,
    ColumnAssociation,
    CompositeColumnCorrelationKey,
)
from metricflow.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow.specs import (
    MetricSpec,
    MeasureSpec,
    DimensionSpec,
    TimeDimensionSpec,
    IdentifierSpec,
    ColumnAssociationResolver,
)
from metricflow.model.semantic_model import SemanticModel
from metricflow.time.time_granularity import TimeGranularity

logger = logging.getLogger(__name__)


class DefaultColumnAssociationResolver(ColumnAssociationResolver):
    """Implements the ColumnAssociationResolver."""

    def __init__(self, semantic_model: SemanticModel) -> None:  # noqa: D
        self._semantic_model = semantic_model

    def resolve_metric_spec(self, metric_spec: MetricSpec) -> ColumnAssociation:  # noqa: D
        return ColumnAssociation(
            column_name=metric_spec.element_name,
            single_column_correlation_key=SingleColumnCorrelationKey(),
        )

    def resolve_measure_spec(self, measure_spec: MeasureSpec) -> ColumnAssociation:  # noqa: D
        return ColumnAssociation(
            column_name=measure_spec.element_name,
            single_column_correlation_key=SingleColumnCorrelationKey(),
        )

    def resolve_dimension_spec(self, dimension_spec: DimensionSpec) -> ColumnAssociation:  # noqa: D
        return ColumnAssociation(
            column_name=StructuredLinkableSpecName(
                identifier_link_names=tuple(x.element_name for x in dimension_spec.identifier_links),
                element_name=dimension_spec.element_name,
            ).qualified_name,
            single_column_correlation_key=SingleColumnCorrelationKey(),
        )

    def resolve_time_dimension_spec(  # noqa: D
        self, time_dimension_spec: TimeDimensionSpec, aggregation_state: Optional[AggregationState] = None
    ) -> ColumnAssociation:
        if time_dimension_spec.time_granularity == TimeGranularity.DAY:
            column_name = StructuredLinkableSpecName(
                identifier_link_names=tuple(x.element_name for x in time_dimension_spec.identifier_links),
                element_name=time_dimension_spec.element_name,
            ).qualified_name
        else:
            column_name = StructuredLinkableSpecName(
                identifier_link_names=tuple(x.element_name for x in time_dimension_spec.identifier_links),
                element_name=time_dimension_spec.element_name,
                time_granularity=time_dimension_spec.time_granularity,
            ).qualified_name

        return ColumnAssociation(
            column_name=column_name + (f"__{aggregation_state.value.lower()}" if aggregation_state else ""),
            single_column_correlation_key=SingleColumnCorrelationKey(),
        )

    def resolve_identifier_spec(self, identifier_spec: IdentifierSpec) -> Tuple[ColumnAssociation, ...]:  # noqa: D
        sub_id_references = []
        for data_source in self._semantic_model.user_configured_model.data_sources:
            for identifier in data_source.identifiers:
                if identifier.reference.element_name == identifier_spec.element_name:
                    sub_id_references = [sub_id.reference for sub_id in identifier.identifiers]
                    break

        # composite identifier case
        if len(sub_id_references) != 0:
            column_associations: Tuple[ColumnAssociation, ...] = ()
            for sub_id_reference in sub_id_references:
                if sub_id_reference is not None:
                    sub_id_name = f"{identifier_spec.element_name}___{sub_id_reference.element_name}"
                    sub_identifier = StructuredLinkableSpecName(
                        identifier_link_names=tuple(x.element_name for x in identifier_spec.identifier_links),
                        element_name=sub_id_name,
                    ).qualified_name
                    column_associations += (
                        ColumnAssociation(
                            column_name=sub_identifier,
                            composite_column_correlation_key=CompositeColumnCorrelationKey(
                                sub_identifier=StructuredLinkableSpecName(
                                    identifier_link_names=(),
                                    element_name=sub_id_name,
                                ).qualified_name
                            ),
                        ),
                    )
            return column_associations

        return (
            ColumnAssociation(
                column_name=StructuredLinkableSpecName(
                    identifier_link_names=tuple(x.element_name for x in identifier_spec.identifier_links),
                    element_name=identifier_spec.element_name,
                ).qualified_name,
                single_column_correlation_key=SingleColumnCorrelationKey(),
            ),
        )



from __future__ import annotations

from typing import List, Optional

from metricflow.dataflow.sql_table import SqlTable
from metricflow.model.objects.common import Metadata
from metricflow.model.objects.base import ModelWithMetadataParsing, HashableBaseModel
from metricflow.object_utils import ExtendedEnum


class MaterializationLocation(ExtendedEnum):
    """Possible locations for a materialized table"""

    DW = "dw"
    FAST_CACHE = "fast_cache"
    TABLEAU = "tableau"


class MaterializationFormat(ExtendedEnum):
    """Possible formats for a materialized table"""

    WIDE = "wide"


class MaterializationTableauParams(HashableBaseModel):
    """Describes the projects to write to in Tableau."""

    projects: List[str]


class MaterializationDestination(HashableBaseModel):
    """Describes where/how a materialized table should be written"""

    location: MaterializationLocation
    format: MaterializationFormat
    rollups: Optional[List[List[str]]]
    tableau_params: Optional[MaterializationTableauParams]


class Materialization(HashableBaseModel, ModelWithMetadataParsing):
    """Describes a materialization"""

    name: str
    description: Optional[str]
    metrics: List[str]
    dimensions: List[str]
    destinations: Optional[List[MaterializationDestination]]
    destination_table: Optional[SqlTable]
    metadata: Optional[Metadata]



from __future__ import annotations

import datetime
from dataclasses import dataclass
from typing import Optional

import pandas as pd

from metricflow.dataclass_serialization import SerializableDataclass
from metricflow.time.time_granularity import TimeGranularity


@dataclass(frozen=True)
class TimeRangeConstraint(SerializableDataclass):
    """Describes how the time dimension for metrics should be constrained."""

    start_time: datetime.datetime
    end_time: datetime.datetime

    @staticmethod
    def ALL_TIME_BEGIN() -> datetime.datetime:  # noqa: D
        return datetime.datetime(2000, 1, 1)

    @staticmethod
    def ALL_TIME_END() -> datetime.datetime:  # noqa: D
        return datetime.datetime(2040, 12, 31)

    @staticmethod
    def all_time() -> TimeRangeConstraint:
        """Return the range representing all time.

        This could also be represented with None as the ends, but doing this makes the logic simpler in many cases.
        """
        return TimeRangeConstraint(
            start_time=TimeRangeConstraint.ALL_TIME_BEGIN(),
            end_time=TimeRangeConstraint.ALL_TIME_END(),
        )

    @staticmethod
    def empty_time() -> TimeRangeConstraint:
        """Return the range representing no time"""
        return TimeRangeConstraint(
            start_time=TimeRangeConstraint.ALL_TIME_BEGIN(),
            end_time=TimeRangeConstraint.ALL_TIME_BEGIN(),
        )

    def _adjust_time_constraint_start_by_window(
        self,
        time_granularity: TimeGranularity,
        time_unit_count: int,
    ) -> TimeRangeConstraint:
        """Moves the start of the time constraint back by 1 window

        if the metric is weekly-active-users (ie window = 1 week) it moves time_constraint.start one week earlier
        """
        start_ts = pd.Timestamp(self.start_time)
        offset = time_granularity.offset_period * time_unit_count
        adjusted_start = (start_ts - offset).to_pydatetime()
        return TimeRangeConstraint(
            start_time=adjusted_start,
            end_time=self.end_time,
        )

    def adjust_time_constraint_for_cumulative_metric(
        self, granularity: Optional[TimeGranularity], count: int
    ) -> TimeRangeConstraint:
        """Given a time constraint for the overall query, adjust it to cover the time range for this metric."""
        if granularity is not None:
            return self._adjust_time_constraint_start_by_window(granularity, count - 1)

        # if no window is specified we want to accumulate from the beginning of time
        return TimeRangeConstraint(
            start_time=TimeRangeConstraint.ALL_TIME_BEGIN(),
            end_time=self.end_time,
        )

    def is_subset_of(self, other: TimeRangeConstraint) -> bool:  # noqa: D
        return self.start_time >= other.start_time and self.end_time <= other.end_time

    def __str__(self) -> str:  # noqa: D
        return f"[{self.start_time.isoformat()}, {self.end_time.isoformat()}]"

    def __repr__(self) -> str:  # noqa: D
        return (
            f"{self.__class__.__name__}(start_time='{self.start_time.isoformat()}', "
            f"end_time='{self.end_time.isoformat()}')"
        )

    def intersection(self, other: TimeRangeConstraint) -> TimeRangeConstraint:  # noqa: D
        # self is completely before the other
        if self.end_time < other.start_time:
            return TimeRangeConstraint.empty_time()
        # self starts before the start of other, and self ends within other
        elif self.start_time <= other.start_time <= self.end_time <= other.end_time:
            return TimeRangeConstraint(
                start_time=other.start_time,
                end_time=self.end_time,
            )
        # self starts before the start of other, and self ends after other
        elif self.start_time <= other.start_time <= other.end_time <= self.end_time:
            return other
        # self starts after the start of other, and self ends within other:
        elif other.start_time <= self.start_time <= self.end_time <= other.end_time:
            return self
        # self starts after the start of other, and self ends after other:
        elif other.start_time <= self.start_time <= other.end_time <= self.end_time:
            return TimeRangeConstraint(
                start_time=self.start_time,
                end_time=other.end_time,
            )
        # self is completely after other
        elif self.start_time > other.end_time:
            return TimeRangeConstraint.empty_time()
        else:
            raise RuntimeError(f"Unhandled case - self: {self} other: {other}")



import datetime
import logging
from dataclasses import dataclass

import _pytest.config
import pytest
from _pytest.fixtures import FixtureRequest

from metricflow.configuration.env_var import EnvironmentVariable
from metricflow.object_utils import random_id

logger = logging.getLogger(__name__)


@dataclass
class MetricFlowTestSessionState:
    """State that is shared between tests during a testing session."""

    sql_engine_url: str
    sql_engine_password: str
    # Where MF system tables should be stored.
    mf_system_schema: str
    # Where tables for test data sets should be stored.
    mf_source_schema: str

    # Number of plans that were displayed to the user.
    display_plans: bool
    # Whether to overwrite any text files that were generated.
    overwrite_snapshots: bool
    # Number of plans that were displayed to the user.
    plans_displayed: int
    # Maximum number of plans to display to the user. If this is exceeded, an exception should be thrown. This is to
    # help avoid a case where an excessive number of plans are displayed. This could happen if the user accidentally
    # runs all tests when they were just looking to run one test and visualize the associated plan.
    max_plans_displayed: int


DISPLAY_PLANS_CLI_FLAG = "--display-plans"
OVERWRITE_SNAPSHOTS_CLI_FLAG = "--overwrite-snapshots"


def pytest_addoption(parser: _pytest.config.argparsing.Parser) -> None:
    """Add options for running pytest through the CLI."""
    parser.addoption(DISPLAY_PLANS_CLI_FLAG, action="store_true", help="Displays plans as SVGs in a browser tab if set")
    parser.addoption(
        OVERWRITE_SNAPSHOTS_CLI_FLAG,
        action="store_true",
        help="Overwrites existing snapshots by ones generated during this testing session",
    )


class MetricFlowTestEnvironmentVariables:
    """Environment variables to setup the testing environment."""

    MF_SQL_ENGINE_URL = EnvironmentVariable("MF_SQL_ENGINE_URL")
    MF_SQL_ENGINE_PASSWORD = EnvironmentVariable("MF_SQL_ENGINE_PASSWORD")


@pytest.fixture(scope="session")
def mf_test_session_state(request: FixtureRequest) -> MetricFlowTestSessionState:  # noqa: D
    engine_url = MetricFlowTestEnvironmentVariables.MF_SQL_ENGINE_URL.get_optional()
    if engine_url is None:
        logger.info(f"{MetricFlowTestEnvironmentVariables.MF_SQL_ENGINE_URL.name} has not been set, so using DuckDb")
        engine_url = "duckdb://"
    engine_password = MetricFlowTestEnvironmentVariables.MF_SQL_ENGINE_PASSWORD.get_optional() or ""

    current_time = datetime.datetime.now().strftime("%Y_%m_%d")
    random_suffix = random_id()
    mf_system_schema = f"mf_test_{current_time}_{random_suffix}"
    mf_source_schema = mf_system_schema

    return MetricFlowTestSessionState(
        sql_engine_url=engine_url,
        sql_engine_password=engine_password,
        mf_system_schema=mf_system_schema,
        mf_source_schema=mf_source_schema,
        display_plans=bool(request.config.getoption(DISPLAY_PLANS_CLI_FLAG, default=False)),
        overwrite_snapshots=bool(request.config.getoption(OVERWRITE_SNAPSHOTS_CLI_FLAG, default=False)),
        plans_displayed=0,
        max_plans_displayed=6,
    )



import logging
from typing import Sequence

import pytest

from metricflow.model.semantic_model import SemanticModel
from metricflow.model.semantics.linkable_spec_resolver import (
    ValidLinkableSpecResolver,
    LinkableElementProperties,
)
from metricflow.model.semantics.semantic_containers import MAX_JOIN_HOPS
from metricflow.references import MetricReference

logger = logging.getLogger(__name__)


@pytest.fixture
def simple_model_spec_resolver(simple_semantic_model: SemanticModel) -> ValidLinkableSpecResolver:  # noqa: D
    return ValidLinkableSpecResolver(
        user_configured_model=simple_semantic_model.user_configured_model,
        max_identifier_links=MAX_JOIN_HOPS,
    )


def test_linkable_spec_resolver(simple_model_spec_resolver: ValidLinkableSpecResolver) -> None:  # noqa: D
    result = simple_model_spec_resolver.get_linkable_elements_for_metrics(
        metric_references=[MetricReference(element_name="bookings"), MetricReference(element_name="views")],
        with_any_of=LinkableElementProperties.all_properties(),
        without_any_of=frozenset({LinkableElementProperties.DERIVED_TIME_GRANULARITY}),
    ).as_spec_set

    assert [
        "create_a_cycle_in_the_join_graph__is_instant",
        "create_a_cycle_in_the_join_graph__listing__capacity_latest",
        "create_a_cycle_in_the_join_graph__listing__country_latest",
        "create_a_cycle_in_the_join_graph__listing__is_lux_latest",
        "listing__capacity_latest",
        "listing__country_latest",
        "listing__is_lux_latest",
        "listing__user__company_name",
        "listing__user__home_state",
        "listing__user__home_state_latest",
    ] == sorted(tuple(x.qualified_name for x in result.dimension_specs))
    assert [
        "create_a_cycle_in_the_join_graph__booking_paid_at",
        "create_a_cycle_in_the_join_graph__listing__created_at",
        "create_a_cycle_in_the_join_graph__listing__ds",
        "ds",
        "ds_partitioned",
        "listing__created_at",
        "listing__ds",
        "listing__user__created_at",
        "listing__user__ds_partitioned",
    ] == sorted(tuple(x.qualified_name for x in result.time_dimension_specs))
    assert [
        "create_a_cycle_in_the_join_graph",
        "create_a_cycle_in_the_join_graph__listing",
        "create_a_cycle_in_the_join_graph__listing__lux_listing",
        "create_a_cycle_in_the_join_graph__listing__user",
        "listing",
        "listing__lux_listing",
        "listing__user",
        "listing__user__company",
    ] == sorted(tuple(x.qualified_name for x in result.identifier_specs))


def property_check_helper(  # noqa: D
    spec_resolver: ValidLinkableSpecResolver,
    metric_references: Sequence[MetricReference],
    element_property: LinkableElementProperties,
    expected_names: Sequence[str],
) -> None:
    results = spec_resolver.get_linkable_elements_for_metrics(
        metric_references=metric_references,
        with_any_of=frozenset({element_property}),
        without_any_of=frozenset(),
    ).as_spec_set.as_tuple

    actual_names = sorted(tuple(x.qualified_name for x in results))
    assert expected_names == actual_names


def test_local_property(simple_model_spec_resolver: ValidLinkableSpecResolver) -> None:  # noqa: D
    property_check_helper(
        spec_resolver=simple_model_spec_resolver,
        metric_references=[MetricReference(element_name="listings")],
        element_property=LinkableElementProperties.LOCAL,
        expected_names=[
            "capacity_latest",
            "country_latest",
            "created_at",
            "created_at__month",
            "created_at__quarter",
            "created_at__week",
            "created_at__year",
            "ds",
            "ds__month",
            "ds__quarter",
            "ds__week",
            "ds__year",
            "is_lux_latest",
            "listing",
            "listing__capacity_latest",
            "listing__country_latest",
            "listing__created_at",
            "listing__created_at__month",
            "listing__created_at__quarter",
            "listing__created_at__week",
            "listing__created_at__year",
            "listing__ds",
            "listing__ds__month",
            "listing__ds__quarter",
            "listing__ds__week",
            "listing__ds__year",
            "listing__is_lux_latest",
            "user",
        ],
    )


def test_local_linked_property(simple_model_spec_resolver: ValidLinkableSpecResolver) -> None:  # noqa: D
    property_check_helper(
        spec_resolver=simple_model_spec_resolver,
        metric_references=[MetricReference(element_name="listings")],
        element_property=LinkableElementProperties.LOCAL_LINKED,
        expected_names=[
            "listing__capacity_latest",
            "listing__country_latest",
            "listing__created_at",
            "listing__created_at__month",
            "listing__created_at__quarter",
            "listing__created_at__week",
            "listing__created_at__year",
            "listing__ds",
            "listing__ds__month",
            "listing__ds__quarter",
            "listing__ds__week",
            "listing__ds__year",
            "listing__is_lux_latest",
        ],
    )


def test_joined_property(simple_model_spec_resolver: ValidLinkableSpecResolver) -> None:  # noqa: D
    property_check_helper(
        spec_resolver=simple_model_spec_resolver,
        metric_references=[MetricReference(element_name="listings")],
        element_property=LinkableElementProperties.JOINED,
        expected_names=[
            "listing__lux_listing",
            "user__company",
            "user__company_name",
            "user__created_at",
            "user__created_at__month",
            "user__created_at__quarter",
            "user__created_at__week",
            "user__created_at__year",
            "user__ds_partitioned",
            "user__ds_partitioned__month",
            "user__ds_partitioned__quarter",
            "user__ds_partitioned__week",
            "user__ds_partitioned__year",
            "user__home_state",
            "user__home_state_latest",
        ],
    )


def test_multi_hop_property(multi_hop_join_semantic_model: SemanticModel) -> None:  # noqa: D
    multi_hop_spec_resolver = ValidLinkableSpecResolver(
        user_configured_model=multi_hop_join_semantic_model.user_configured_model,
        max_identifier_links=MAX_JOIN_HOPS,
    )
    property_check_helper(
        spec_resolver=multi_hop_spec_resolver,
        metric_references=[MetricReference(element_name="txn_count")],
        element_property=LinkableElementProperties.MULTI_HOP,
        expected_names=[
            "account_id__customer_id__country",
            "account_id__customer_id__customer_atomic_weight",
            "account_id__customer_id__customer_name",
            "account_id__customer_id__customer_third_hop_id",
            "account_id__customer_id__ds_partitioned",
            "account_id__customer_id__ds_partitioned__month",
            "account_id__customer_id__ds_partitioned__quarter",
            "account_id__customer_id__ds_partitioned__week",
            "account_id__customer_id__ds_partitioned__year",
        ],
    )


def test_derived_time_granularity_property(simple_model_spec_resolver: ValidLinkableSpecResolver) -> None:  # noqa: D
    property_check_helper(
        spec_resolver=simple_model_spec_resolver,
        metric_references=[MetricReference(element_name="listings")],
        element_property=LinkableElementProperties.DERIVED_TIME_GRANULARITY,
        expected_names=[
            "created_at__month",
            "created_at__quarter",
            "created_at__week",
            "created_at__year",
            "ds__month",
            "ds__quarter",
            "ds__week",
            "ds__year",
            "listing__created_at__month",
            "listing__created_at__quarter",
            "listing__created_at__week",
            "listing__created_at__year",
            "listing__ds__month",
            "listing__ds__quarter",
            "listing__ds__week",
            "listing__ds__year",
            "user__created_at__month",
            "user__created_at__quarter",
            "user__created_at__week",
            "user__created_at__year",
            "user__ds_partitioned__month",
            "user__ds_partitioned__quarter",
            "user__ds_partitioned__week",
            "user__ds_partitioned__year",
        ],
    )


def test_identifier_property(simple_model_spec_resolver: ValidLinkableSpecResolver) -> None:  # noqa: D
    property_check_helper(
        spec_resolver=simple_model_spec_resolver,
        metric_references=[MetricReference(element_name="listings")],
        element_property=LinkableElementProperties.IDENTIFIER,
        expected_names=["listing", "listing__lux_listing", "user", "user__company"],
    )



import os
import yaml

from typing import Dict, Optional


class YamlFileHandler:
    """Class to handle interactions with a non-nested yaml."""

    def __init__(self, yaml_file_path: str) -> None:  # noqa: D
        self.yaml_file_path = yaml_file_path

    def _load_yaml(self) -> Dict[str, str]:
        """Reads the provided yaml file and loads it into a dictionary."""
        content: Dict[str, str] = {}
        if os.path.exists(self.yaml_file_path):
            with open(self.yaml_file_path) as f:
                content = yaml.load(f, Loader=yaml.SafeLoader) or {}
        return content

    def get_value(self, key: str) -> Optional[str]:
        """Attempts to get a corresponding value from the yaml file. Throw an error if not exists or None."""
        content = self._load_yaml()
        return content.get(key)

    def set_value(self, key: str, value: str) -> None:
        """Sets a value to a given key in yaml file."""
        content = self._load_yaml()
        content[key] = value
        with open(self.yaml_file_path, "w") as f:
            yaml.dump(content, f)

    def remove_value(self, key: str) -> None:
        """Removes a key in yaml file."""
        content = self._load_yaml()
        if key not in content:
            return
        del content[key]
        with open(self.yaml_file_path, "w") as f:
            yaml.dump(content, f)

    @property
    def url(self) -> str:
        """Returns the file url of this handler."""
        return f"file:/{self.yaml_file_path}"



from __future__ import annotations

from pydantic import validator
from typing import Any, Optional, List

from metricflow.model.objects.base import HashableBaseModel, ModelWithMetadataParsing
from metricflow.model.objects.common import Metadata
from metricflow.object_utils import ExtendedEnum
from metricflow.references import IdentifierReference, CompositeSubIdentifierReference


class IdentifierType(ExtendedEnum):
    """Defines uniqueness and completeness of identifier"""

    FOREIGN = "foreign"
    PRIMARY = "primary"
    UNIQUE = "unique"

    # for identifiers that are only rendered, not joined on (used when rendering composite identifiers)
    # RENDER_ONLY = "render_only" # DEPRECATED ? we shouldnt need in the new world?


class CompositeSubIdentifier(HashableBaseModel):
    """CompositeSubIdentifiers either describe or reference the identifiers that comprise a composite identifier"""

    name: Optional[str]
    expr: Optional[str]
    ref: Optional[str]

    @property
    def reference(self) -> CompositeSubIdentifierReference:  # noqa: D
        assert self.name, f"The element name should have been set during model transformation. Got {self}"
        return CompositeSubIdentifierReference(element_name=self.name)


class Identifier(HashableBaseModel, ModelWithMetadataParsing):
    """Describes a identifier"""

    name: str
    description: Optional[str]
    type: IdentifierType
    role: Optional[str]
    entity: Optional[str]
    identifiers: List[CompositeSubIdentifier] = []
    expr: Optional[str] = None
    metadata: Optional[Metadata]

    @validator("entity", always=True)
    @classmethod
    def default_entity_value(cls, value: Any, values: Any) -> str:  # type: ignore[misc]
        """Defaulting the value of the identifier 'entity' value using pydantic validator

        If an entity value is provided that is a string, that will become the value of
        entity. If the provifed entity value is None, the entity value becomes the
        element_name representation of the identifier's name.
        """

        if value is None:
            if "name" not in values:
                raise ValueError("Failed to default entity value because objects name value was not defined")
            value = values["name"]

        # guarantee value is string
        if not isinstance(value, str):
            raise ValueError(f"Entity value should be a string (str) type, but got {type(value)} with value: {value}")
        return value

    @property
    def is_primary_time(self) -> bool:  # noqa: D
        return False

    @property
    def is_composite(self) -> bool:  # noqa: D
        return self.identifiers is not None and len(self.identifiers) > 0

    @property
    def reference(self) -> IdentifierReference:  # noqa: D
        return IdentifierReference(element_name=self.name)



from __future__ import annotations

import logging
from collections import defaultdict
from copy import deepcopy
from typing import Dict, List, Set, Optional, Sequence, Tuple, FrozenSet

from metricflow.errors.errors import (
    DuplicateMetricError,
    MetricNotFoundError,
    NonExistentMeasureError,
    InvalidDataSourceError,
)
from metricflow.instances import DataSourceReference, DataSourceElementReference
from metricflow.aggregation_properties import AggregationType
from metricflow.model.objects.data_source import DataSource, DataSourceOrigin
from metricflow.model.objects.elements.dimension import Dimension
from metricflow.model.objects.elements.identifier import Identifier
from metricflow.model.objects.elements.measure import Measure
from metricflow.model.objects.metric import Metric, MetricType
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.semantics.data_source_container import PydanticDataSourceContainer
from metricflow.model.semantics.element_group import ElementGrouper
from metricflow.model.semantics.linkable_spec_resolver import (
    ValidLinkableSpecResolver,
    LinkableElementProperties,
)
from metricflow.model.spec_converters import MeasureConverter, WhereConstraintConverter
from metricflow.references import (
    DimensionReference,
    IdentifierReference,
    LinkableElementReference,
    MeasureReference,
    MetricReference,
    TimeDimensionReference,
)
from metricflow.specs import (
    LinkableInstanceSpec,
    MeasureSpec,
    MetricInputMeasureSpec,
    NonAdditiveDimensionSpec,
)

logger = logging.getLogger(__name__)

MAX_JOIN_HOPS = 2


class MetricSemantics:  # noqa: D
    def __init__(  # noqa: D
        self, user_configured_model: UserConfiguredModel, data_source_semantics: DataSourceSemantics
    ) -> None:
        self._user_configured_model = user_configured_model
        self._metrics: Dict[MetricReference, Metric] = {}
        self._data_source_semantics = data_source_semantics

        # Dict from the name of the metric to the hash.
        self._metric_hashes: Dict[MetricReference, str] = {}

        for metric in self._user_configured_model.metrics:
            self.add_metric(metric)

        self._linkable_spec_resolver = ValidLinkableSpecResolver(
            user_configured_model=self._user_configured_model,
            max_identifier_links=MAX_JOIN_HOPS,
        )

    def element_specs_for_metrics(
        self,
        metric_references: List[MetricReference],
        with_any_property: FrozenSet[LinkableElementProperties] = LinkableElementProperties.all_properties(),
        without_any_property: FrozenSet[LinkableElementProperties] = frozenset(),
    ) -> List[LinkableInstanceSpec]:
        """Dimensions common to all metrics requested (intersection)"""

        all_linkable_specs = self._linkable_spec_resolver.get_linkable_elements_for_metrics(
            metric_references=metric_references,
            with_any_of=with_any_property,
            without_any_of=without_any_property,
        ).as_spec_set

        return sorted(all_linkable_specs.as_tuple, key=lambda x: x.qualified_name)

    def get_metrics(self, metric_references: List[MetricReference]) -> List[Metric]:  # noqa: D
        res = []
        for metric_reference in metric_references:
            if metric_reference not in self._metrics:
                raise MetricNotFoundError(
                    f"Unable to find metric `{metric_reference}`. Perhaps it has not been registered"
                )
            res.append(self._metrics[metric_reference])

        return res

    @property
    def metric_references(self) -> List[MetricReference]:  # noqa: D
        return list(self._metrics.keys())

    def get_metric(self, metric_reference: MetricReference) -> Metric:  # noqa:D
        if metric_reference not in self._metrics:
            raise MetricNotFoundError(f"Unable to find metric `{metric_reference}`. Perhaps it has not been registered")
        return self._metrics[metric_reference]

    def add_metric(self, metric: Metric) -> None:
        """Add metric, validating presence of required measures"""
        metric_reference = MetricReference(element_name=metric.name)
        if metric_reference in self._metrics:
            raise DuplicateMetricError(f"Metric `{metric.name}` has already been registered")
        for measure_reference in metric.measure_references:
            if measure_reference not in self._data_source_semantics.measure_references:
                raise NonExistentMeasureError(
                    f"Metric `{metric.name}` references measure `{measure_reference}` which has not been registered"
                )
        self._metrics[metric_reference] = metric
        self._metric_hashes[metric_reference] = metric.definition_hash

    @property
    def valid_hashes(self) -> Set[str]:
        """Return all of the hashes of the metric definitions."""
        return set(self._metric_hashes.values())

    def measures_for_metric(self, metric_reference: MetricReference) -> Tuple[MetricInputMeasureSpec, ...]:
        """Return the measure specs required to compute the metric."""
        metric = self.get_metric(metric_reference)
        input_measure_specs: List[MetricInputMeasureSpec] = []

        for input_measure in metric.input_measures:
            spec_constraint = (
                WhereConstraintConverter.convert_to_spec_where_constraint(
                    data_source_semantics=self._data_source_semantics,
                    where_constraint=input_measure.constraint,
                )
                if input_measure.constraint is not None
                else None
            )
            measure_spec = MeasureSpec(
                element_name=input_measure.name,
                non_additive_dimension_spec=self._data_source_semantics.non_additive_dimension_specs_by_measure.get(
                    input_measure.measure_reference
                ),
            )
            spec = MetricInputMeasureSpec(
                measure_spec=measure_spec,
                constraint=spec_constraint,
                alias=input_measure.alias,
            )
            input_measure_specs.append(spec)

        return tuple(input_measure_specs)

    def contains_cumulative_metric(self, metric_references: Sequence[MetricReference]) -> bool:
        """Returns true if any of the specs correspond to a cumulative metric."""
        for metric_reference in metric_references:
            if self.get_metric(metric_reference).type == MetricType.CUMULATIVE:
                return True
        return False


class DataSourceSemantics:
    """Tracks semantic information for data source held in a set of DataSourceContainers

    This implements both the DataSourceSemanticsAccessors protocol, the interface type we use throughout the codebase.
    That interface prevents unwanted calls to methods for adding data sources to the container.
    """

    def __init__(  # noqa: D
        self,
        model: UserConfiguredModel,
        configured_data_source_container: PydanticDataSourceContainer,
    ) -> None:
        self._model = model
        self._measure_index: Dict[MeasureReference, List[DataSource]] = defaultdict(list)
        self._measure_aggs: Dict[
            MeasureReference, AggregationType
        ] = {}  # maps measures to their one consistent aggregation
        self._measure_agg_time_dimension: Dict[MeasureReference, TimeDimensionReference] = {}
        self._measure_non_additive_dimension_specs: Dict[MeasureReference, NonAdditiveDimensionSpec] = {}
        self._dimension_index: Dict[DimensionReference, List[DataSource]] = defaultdict(list)
        self._linkable_reference_index: Dict[LinkableElementReference, List[DataSource]] = defaultdict(list)
        self._entity_index: Dict[Optional[str], List[DataSource]] = defaultdict(list)
        self._identifier_ref_to_entity: Dict[IdentifierReference, Optional[str]] = {}
        self._data_source_names: Set[str] = set()

        self._configured_data_source_container = configured_data_source_container
        self._data_source_to_aggregation_time_dimensions: Dict[
            DataSourceReference, ElementGrouper[TimeDimensionReference, MeasureSpec]
        ] = {}

        # Add semantic tracking for data sources from configured_data_source_container
        for data_source in self._configured_data_source_container.values():
            assert isinstance(data_source, DataSource)
            self.add_configured_data_source(data_source)

    def add_configured_data_source(self, data_source: DataSource) -> None:
        """Dont use this unless you mean it (ie in tests). The configured data sources are supposed to be static"""
        self._configured_data_source_container._put(data_source)
        self._add_data_source(data_source)

    def get_dimension_references(self) -> List[DimensionReference]:  # noqa: D
        return list(self._dimension_index.keys())

    def get_dimension(
        self, dimension_reference: DimensionReference, origin: Optional[DataSourceOrigin] = None
    ) -> Dimension:
        """Retrieves a full dimension object by name"""
        for dimension_source in self._dimension_index[dimension_reference]:
            if origin and dimension_source.origin != origin:
                continue
            dimension = dimension_source.get_dimension(dimension_reference)
            # find the data source that has the requested dimension by the requested identifier

            return deepcopy(dimension)

        raise ValueError(
            f"Could not find dimension with name ({dimension_reference.element_name}) in configured data sources"
        )

    def get_time_dimension(self, time_dimension_reference: TimeDimensionReference) -> Dimension:
        """Retrieves a full dimension object by name"""
        dimension_reference = time_dimension_reference.dimension_reference()

        if dimension_reference not in self._dimension_index:
            raise ValueError(
                f"Could not find dimension with name ({dimension_reference.element_name}) in configured data sources"
            )

        for dimension_source in self._dimension_index[dimension_reference]:
            dimension = dimension_source.get_dimension(dimension_reference)
            # TODO: Unclear if the deepcopy is necessary.
            return deepcopy(dimension)

        assert False, f"{time_dimension_reference} should have been in the dimension index"

    @property
    def measure_references(self) -> List[MeasureReference]:  # noqa: D
        return list(self._measure_index.keys())

    @property
    def non_additive_dimension_specs_by_measure(self) -> Dict[MeasureReference, NonAdditiveDimensionSpec]:  # noqa: D
        return self._measure_non_additive_dimension_specs

    def get_measure(self, measure_reference: MeasureReference) -> Measure:  # noqa: D
        if measure_reference not in self._measure_index:
            raise ValueError(f"Could not find measure with name ({measure_reference}) in configured data sources")

        assert len(self._measure_index[measure_reference]) >= 1
        # Measures should be consistent across data sources, so just use the first one.
        return list(self._measure_index[measure_reference])[0].get_measure(measure_reference)

    def get_identifier_references(self) -> List[IdentifierReference]:  # noqa: D
        return list(self._identifier_ref_to_entity.keys())

    # DSC interface
    def get_data_sources_for_measure(self, measure_reference: MeasureReference) -> List[DataSource]:  # noqa: D
        return self._measure_index[measure_reference]

    def get_agg_time_dimension_for_measure(  # noqa: D
        self, measure_reference: MeasureReference
    ) -> TimeDimensionReference:
        return self._measure_agg_time_dimension[measure_reference]

    def get_identifier_in_data_source(self, ref: DataSourceElementReference) -> Optional[Identifier]:  # Noqa: d
        data_source = self.get(ref.data_source_name)
        if not data_source:
            return None

        for identifier in data_source.identifiers:
            if identifier.reference.element_name == ref.element_name:
                return identifier

        return None

    def get(self, data_source_name: str) -> Optional[DataSource]:  # noqa: D
        if data_source_name in self._configured_data_source_container:
            data_source = self._configured_data_source_container.get(data_source_name)
            assert isinstance(data_source, DataSource)
            return data_source

        return None

    def get_by_reference(self, data_source_reference: DataSourceReference) -> Optional[DataSource]:  # noqa: D
        return self.get(data_source_reference.data_source_name)

    def _add_data_source(
        self,
        data_source: DataSource,
        fail_on_error: bool = True,
        logging_context: str = "",
    ) -> None:
        """Add data source semantic information, validating consistency with existing data sources."""
        errors = []

        if data_source.name in self._data_source_names:
            errors.append(f"name {data_source.name} already registered - please ensure data source names are unique")

        for measure in data_source.measures:
            if measure.reference in self._measure_aggs and self._measure_aggs[measure.reference] != measure.agg:
                errors.append(
                    f"conflicting aggregation (agg) for measure `{measure.reference.element_name}` registered as "
                    f"`{self._measure_aggs[measure.reference]}`; Got `{measure.agg}"
                )

        if errors:
            error_prefix = "\n  - "
            error_msg = (
                f"Unable to add data source `{data_source.name}` "
                f"{'while ' + logging_context + ' ' if logging_context else ''}"
                f"{'... skipping' if not fail_on_error else ''}.\n"
                f"Errors: {error_prefix + error_prefix.join(errors)}"
            )
            if fail_on_error:
                raise InvalidDataSourceError(error_msg)
            logger.warning(error_msg)
            return

        self._data_source_names.add(data_source.name)
        self._data_source_to_aggregation_time_dimensions[data_source.reference] = ElementGrouper[
            TimeDimensionReference, MeasureSpec
        ]()

        for measure in data_source.measures:
            self._measure_aggs[measure.reference] = measure.agg
            self._measure_index[measure.reference].append(data_source)
            agg_time_dimension = measure.checked_agg_time_dimension
            self._data_source_to_aggregation_time_dimensions[data_source.reference].add_value(
                key=agg_time_dimension,
                value=MeasureConverter.convert_to_measure_spec(measure=measure),
            )
            self._measure_agg_time_dimension[measure.reference] = agg_time_dimension
            if measure.non_additive_dimension:
                non_additive_dimension_spec = NonAdditiveDimensionSpec(
                    name=measure.non_additive_dimension.name,
                    window_choice=measure.non_additive_dimension.window_choice,
                    window_groupings=tuple(measure.non_additive_dimension.window_groupings),
                )
                self._measure_non_additive_dimension_specs[measure.reference] = non_additive_dimension_spec
        for dim in data_source.dimensions:
            self._linkable_reference_index[dim.reference].append(data_source)
            self._dimension_index[dim.reference].append(data_source)
        for ident in data_source.identifiers:
            self._identifier_ref_to_entity[ident.reference] = ident.entity
            self._entity_index[ident.entity].append(data_source)
            self._linkable_reference_index[ident.reference].append(data_source)

    @property
    def data_source_references(self) -> Sequence[DataSourceReference]:  # noqa: D
        data_source_names_sorted = sorted(self._data_source_names)
        return tuple(DataSourceReference(data_source_name=x) for x in data_source_names_sorted)

    def get_aggregation_time_dimensions_with_measures(
        self, data_source_reference: DataSourceReference
    ) -> ElementGrouper[TimeDimensionReference, MeasureSpec]:
        """Return all time dimensions in a data source with their associated measures."""
        assert (
            data_source_reference in self._data_source_to_aggregation_time_dimensions
        ), f"Data Source {data_source_reference} is not known"
        return self._data_source_to_aggregation_time_dimensions[data_source_reference]

    def get_data_sources_for_identifier(self, identifier_reference: IdentifierReference) -> Set[DataSource]:
        """Return all data sources associated with an identifier reference"""
        identifier_entity = self._identifier_ref_to_entity[identifier_reference]
        return set(self._entity_index[identifier_entity])



import pytest

from metricflow.dataset.dataset import DataSet
from metricflow.model.model_validator import ModelValidator
from metricflow.model.validations.materializations import ValidMaterializationRule
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.test.model.validations.helpers import materialization_with_guaranteed_meta
from metricflow.test.test_utils import model_with_materialization


def test_can_configure_model_validator_rules(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa: D
    model = model_with_materialization(
        simple_model__pre_transforms,
        [
            materialization_with_guaranteed_meta(
                name="foobar",
                metrics=["invalid_bookings"],
                dimensions=[DataSet.metric_time_dimension_name()],
            )
        ],
    )

    # confirm that with the default configuration, an issue is raised
    issues = ModelValidator().validate_model(model).issues
    assert len(issues.all_issues) == 1, f"ModelValidator with default rules had unexpected number of issues {issues}"

    # confirm that a custom configuration excluding ValidMaterializationRule, no issue is raised
    rules = [rule for rule in ModelValidator.DEFAULT_RULES if rule.__class__ is not ValidMaterializationRule]
    issues = ModelValidator(rules=rules).validate_model(model).issues
    assert len(issues.all_issues) == 0, f"ModelValidator without ValidMaterializationRule returned issues {issues}"


def test_cant_configure_model_validator_without_rules() -> None:  # noqa: D
    with pytest.raises(ValueError):
        ModelValidator(rules=[])

    with pytest.raises(ValueError):
        ModelValidator(rules=())

    with pytest.raises(ValueError):
        ModelValidator(rules=None)  # type: ignore



import logging
from typing import List
from metricflow.instances import DataSourceElementReference, DataSourceReference

from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.elements.dimension import DimensionType
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.validator_helpers import (
    DataSourceContext,
    DataSourceElementContext,
    DataSourceElementType,
    FileContext,
    ModelValidationRule,
    ValidationIssueType,
    ValidationError,
    validate_safely,
)
from metricflow.time.time_constants import SUPPORTED_GRANULARITIES

logger = logging.getLogger(__name__)


class DataSourceTimeDimensionWarningsRule(ModelValidationRule):
    """Checks time dimensions in data sources."""

    @staticmethod
    @validate_safely(whats_being_done="running model validation ensuring time dimensions are defined properly")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        issues: List[ValidationIssueType] = []

        for data_source in model.data_sources:
            issues.extend(DataSourceTimeDimensionWarningsRule._validate_data_source(data_source=data_source))
        return issues

    @staticmethod
    @validate_safely(whats_being_done="checking validity of the data source's time dimensions")
    def _validate_data_source(data_source: DataSource) -> List[ValidationIssueType]:
        issues: List[ValidationIssueType] = []

        primary_time_dimensions = []

        for dim in data_source.dimensions:
            context = DataSourceElementContext(
                file_context=FileContext.from_metadata(metadata=data_source.metadata),
                data_source_element=DataSourceElementReference(
                    data_source_name=data_source.name, element_name=dim.name
                ),
                element_type=DataSourceElementType.DIMENSION,
            )

            if dim.type == DimensionType.TIME:
                if dim.type_params is None:
                    continue
                elif dim.type_params.is_primary:
                    primary_time_dimensions.append(dim)
                elif dim.type_params.time_granularity:
                    if dim.type_params.time_granularity not in SUPPORTED_GRANULARITIES:
                        issues.append(
                            ValidationError(
                                context=context,
                                message=f"Unsupported time granularity in time dimension with name: {dim.name}, "
                                f"Please use {[s.value for s in SUPPORTED_GRANULARITIES]}",
                            )
                        )

        if len(primary_time_dimensions) == 0 and len(data_source.measures) > 0:
            issues.append(
                ValidationError(
                    context=DataSourceContext(
                        file_context=FileContext.from_metadata(metadata=data_source.metadata),
                        data_source=DataSourceReference(data_source_name=data_source.name),
                    ),
                    message=f"No primary time dimension in data source with name ({data_source.name}). Please add one",
                )
            )

        if len(primary_time_dimensions) > 1:
            for primary_time_dimension in primary_time_dimensions:
                issues.append(
                    ValidationError(
                        context=DataSourceContext(
                            file_context=FileContext.from_metadata(metadata=data_source.metadata),
                            data_source=DataSourceReference(data_source_name=data_source.name),
                        ),
                        message=f"In data source {data_source.name}, "
                        f"Primary time dimension with name: {primary_time_dimension.name} "
                        f"is one of many defined as primary.",
                    )
                )

        return issues



import logging
import os
from dataclasses import dataclass
from string import Template
import traceback
import git
from typing import Optional, Dict, List, Union, Type

from jsonschema import exceptions

from metricflow.errors.errors import ParsingException
from metricflow.model.model_transformer import ModelTransformer
from metricflow.model.objects.common import Version, YamlConfigFile
from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.materialization import Materialization
from metricflow.model.objects.metric import Metric
from metricflow.model.parsing.schemas_internal import (
    metric_validator,
    data_source_validator,
    materialization_validator,
)
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.parsing.yaml_loader import (
    ParsingContext,
    YamlConfigLoader,
    PARSING_CONTEXT_KEY,
)
from metricflow.model.validations.validator_helpers import (
    FileContext,
    ModelValidationException,
    ModelValidationResults,
    ValidationError,
    ValidationIssueType,
)

logger = logging.getLogger(__name__)

VERSION_KEY = "mf_config_schema"
METRIC_TYPE = "metric"
DATA_SOURCE_TYPE = "data_source"
MATERIALIZATION_TYPE = "materialization"
DOCUMENT_TYPES = [METRIC_TYPE, DATA_SOURCE_TYPE, MATERIALIZATION_TYPE]


@dataclass(frozen=True)
class ModelBuildResult:  # noqa: D
    model: UserConfiguredModel
    # Issues found in the model.
    issues: ModelValidationResults = ModelValidationResults()


@dataclass(frozen=True)
class FileParsingResult:
    """Results of parsing a config file

    Attributes:
        elements: MetricFlow model elements parsed from the file
        issues: Issues found when trying to parse the file
    """

    elements: List[Union[DataSource, Metric, Materialization]]
    issues: List[ValidationIssueType]


def collect_yaml_config_file_paths(directory: str) -> List[str]:
    """Collects a list of file paths for model config files

    Ignores files that are:
        - In hidden directories (i.e. directories starting with '.')
        - Hidden files (i.e. files starting with '.')
        - Non YAML files
        - Ignored by the repo's .gitignore file (if a repo is detected)

    NOTE: We ignore files ignored by .gitignore because an issue cropped up wherein
    sometimes dependencies of projects include YAML files in their package.
    """
    config_file_paths: List[str] = []
    for root, dirs, files in os.walk(directory):
        # Skip hidden directories. os.walk() supports mutation of dirs to skip directories.
        dirs[:] = [d for d in dirs if not d.startswith(".")]

        for file in files:
            if not YamlConfigLoader.is_valid_yaml_file_ending(file):
                continue
            # Skip hidden files
            if file.startswith("."):
                continue

            file_path = os.path.join(root, file)
            config_file_paths.append(file_path)

    try:
        repo = git.Repo(directory, search_parent_directories=True)
        # repo.ignored returns a list of file paths which are the file paths
        # that should be ignored as a subset of the handed in file paths
        ignored_files = repo.ignored(config_file_paths)
        config_file_paths = list(set(config_file_paths) - set(ignored_files))
    except git.exc.InvalidGitRepositoryError:
        pass

    return config_file_paths


def parse_directory_of_yaml_files_to_model(
    directory: str,
    template_mapping: Optional[Dict[str, str]] = None,
    apply_pre_transformations: Optional[bool] = True,
    apply_post_transformations: Optional[bool] = True,
    raise_issues_as_exceptions: bool = True,
) -> ModelBuildResult:
    """Parse files in the given directory to a UserConfiguredModel.

    Strings in the file following the Python string template format are replaced according to the template_mapping dict.
    """
    file_paths = collect_yaml_config_file_paths(directory=directory)
    return parse_yaml_file_paths_to_model(
        file_paths=file_paths,
        template_mapping=template_mapping,
        apply_pre_transformations=apply_pre_transformations,
        apply_post_transformations=apply_post_transformations,
        raise_issues_as_exceptions=raise_issues_as_exceptions,
    )


def parse_yaml_file_paths_to_model(
    file_paths: List[str],
    template_mapping: Optional[Dict[str, str]] = None,
    apply_pre_transformations: Optional[bool] = True,
    apply_post_transformations: Optional[bool] = True,
    raise_issues_as_exceptions: bool = True,
) -> ModelBuildResult:
    """Parse files the given list of file paths to a UserConfiguredModel.

    Strings in the files following the Python string template format are replaced according to the template_mapping dict.
    """
    template_mapping = template_mapping or {}
    yaml_config_files = []
    for file_path in file_paths:
        try:
            with open(file_path) as f:
                contents = Template(f.read()).substitute(template_mapping)
                yaml_config_files.append(
                    YamlConfigFile(filepath=file_path, contents=contents),
                )
        except UnicodeDecodeError as e:
            # We could alternatively return this as a validation issue, but this
            # exception is hit *before* building the model. Currently the
            # ModelBuildResult guarantees a UserConfiguredModel. We could make
            # UserConfiguredModel optional on ModelBuildResult, but this has
            # undesirable consequences.
            raise Exception(
                f"The content of file `{file_path}` doesn't match the encoding of the file."
                " If you know the encoding the content is in, try resaving the file with that encoding explicitly."
                " Alternatively this error generally arises due to copy and pasted content,"
                " try manually typing up the problem file instead of copy and pasting"
            ) from e

    build_result = parse_yaml_files_to_model(yaml_config_files)
    model = build_result.model
    assert model

    build_issues = build_result.issues
    try:
        if apply_pre_transformations:
            model = ModelTransformer.pre_validation_transform_model(model)

        if apply_post_transformations:
            model = ModelTransformer.post_validation_transform_model(model)
    except Exception as e:
        transformation_issue_results = ModelValidationResults(errors=[ValidationError(message=str(e))])
        build_issues = ModelValidationResults.merge([build_issues, transformation_issue_results])

    if raise_issues_as_exceptions and build_issues.has_blocking_issues:
        raise ModelValidationException(build_issues.all_issues)

    return ModelBuildResult(model=model, issues=build_issues)


def parse_yaml_files_to_model(
    files: List[YamlConfigFile],
    data_source_class: Type[DataSource] = DataSource,
    metric_class: Type[Metric] = Metric,
    materialization_class: Type[Materialization] = Materialization,
) -> ModelBuildResult:
    """Builds UserConfiguredModel from list of config files (as strings).

    Persistent storage connection may be passed to write parsed objects=
    to storage and populate object metadata

    Note: this function does not finalize the model
    """
    data_sources = []
    metrics = []
    materializations: List[Materialization] = []
    valid_object_classes = [data_source_class.__name__, metric_class.__name__, materialization_class.__name__]
    issues: List[ValidationIssueType] = []

    for config_file in files:
        parsing_result = parse_config_yaml(  # parse config file
            config_file,
            data_source_class=data_source_class,
            metric_class=metric_class,
            materialization_class=materialization_class,
        )
        file_issues = parsing_result.issues
        for obj in parsing_result.elements:
            if isinstance(obj, data_source_class):
                data_sources.append(obj)
            elif isinstance(obj, metric_class):
                metrics.append(obj)
            elif isinstance(obj, materialization_class):
                materializations.append(obj)
            else:
                file_issues.append(
                    ValidationError(
                        context=FileContext(file_name=config_file.filepath),
                        message=f"Unexpected model object {obj.__name__}. Expected {valid_object_classes}.",
                    )
                )

        issues += file_issues

    return ModelBuildResult(
        model=UserConfiguredModel(
            data_sources=data_sources,
            materializations=materializations,
            metrics=metrics,
        ),
        issues=ModelValidationResults.from_issues_sequence(issues),
    )


def parse_config_yaml(
    config_yaml: YamlConfigFile,
    data_source_class: Type[DataSource] = DataSource,
    metric_class: Type[Metric] = Metric,
    materialization_class: Type[Materialization] = Materialization,
) -> FileParsingResult:
    """Parses transform config file passed as string - Returns list of model objects"""
    results: List[Union[DataSource, Metric, Materialization]] = []
    ctx: Optional[ParsingContext] = None
    issues: List[ValidationIssueType] = []
    try:
        for config_document in YamlConfigLoader.load_all_with_context(
            name=config_yaml.filepath, contents=config_yaml.contents
        ):
            # The config document can be None if there is nothing but white space between two `---`
            # this isn't really an issue, so lets just swallow it
            if config_document is None:
                continue
            if not isinstance(config_document, dict):
                issues.append(
                    ValidationError(
                        context=FileContext(file_name=config_yaml.filepath),
                        message=f"YAML must be a dict. Got `{type(config_document)}`.",
                    )
                )
                continue

            keys = config_document.keys()

            # This SHOULDN'T ever happen but if it does, we want to know. If this
            # does ever happen, it is likely due to a change in 'load_all_with_context'
            if PARSING_CONTEXT_KEY not in keys:
                raise RuntimeError(
                    f"No parsing context present. Expected key `{PARSING_CONTEXT_KEY}` from the YAML parser."
                )

            ctx = config_document.pop(PARSING_CONTEXT_KEY)
            assert ctx

            if VERSION_KEY in config_document:
                version = Version.parse(config_document.pop(VERSION_KEY))
                major_version = version.major

                if major_version != 0:
                    issues.append(
                        ValidationError(
                            context=FileContext(file_name=ctx.filename, line_number=ctx.start_line),
                            message=f"Unsupported version {version} in config document.",
                        )
                    )

            # Because we've popped the VERSION KEY and PARSING_CONTEXT_KEY, there
            # should only be the base object key remaining
            if len(keys) != 1:
                issues.append(
                    ValidationError(
                        context=FileContext(file_name=ctx.filename, line_number=ctx.start_line),
                        message=f"Document should have one type of key, but has {keys}.",
                    )
                )
                continue

            # retrieve last top-level key as type
            document_type = next(iter(config_document.keys()))
            object_cfg = config_document[document_type]

            try:
                if document_type == METRIC_TYPE:
                    metric_validator.validate(config_document[document_type])
                    results.append(metric_class.parse_obj(object_cfg))
                elif document_type == DATA_SOURCE_TYPE:
                    data_source_validator.validate(config_document[document_type])
                    results.append(data_source_class.parse_obj(object_cfg))
                elif document_type == MATERIALIZATION_TYPE:
                    materialization_validator.validate(config_document[document_type])
                    results.append(materialization_class.parse_obj(object_cfg))
                else:
                    issues.append(
                        ValidationError(
                            context=FileContext(file_name=ctx.filename, line_number=ctx.start_line),
                            message=f"Invalid document type: {document_type}. Expected {DOCUMENT_TYPES}.",
                        )
                    )
            # catches exceptions from jsonschema validator
            except exceptions.ValidationError as e:
                context = FileContext(file_name=ctx.filename, line_number=ctx.start_line)
                issues.append(
                    ValidationError(
                        context=context,
                        message=f"YAML document did not conform to metric spec.\nError: {e}",
                        extra_detail="".join(traceback.format_tb(e.__traceback__)),
                    )
                )
            # ParsingException: catches exceptions from *.parse_obj calls
            # Exception: general exception for a given document. Basicially we
            # don't want an exception on one document to halt checking the rest
            # of the documents
            except (ParsingException, Exception) as e:
                context = FileContext(file_name=ctx.filename, line_number=ctx.start_line)
                issues.append(
                    ValidationError(
                        context=context,
                        message=str(e),
                        extra_detail="".join(traceback.format_tb(e.__traceback__)),
                    )
                )
    # If a runtime error occured, we still want this to break things
    except RuntimeError:
        raise
    # Any other error should be handled as an issue
    except Exception as e:
        context = FileContext(file_name=config_yaml.filepath)
        issues.append(
            ValidationError(context=context, message=str(e), extra_detail="".join(traceback.format_tb(e.__traceback__)))
        )

    return FileParsingResult(elements=results, issues=issues)



import logging
import re
from dataclasses import dataclass
from typing import Optional, List, Tuple, Sequence

from metricflow.aggregation_properties import AggregationState
from metricflow.dag.id_generation import IdGeneratorRegistry
from metricflow.dataflow.sql_table import SqlTable
from metricflow.dataset.data_source_adapter import DataSourceDataSet
from metricflow.instances import (
    MeasureInstance,
    DataSourceElementReference,
    DimensionInstance,
    TimeDimensionInstance,
    IdentifierInstance,
    InstanceSet,
    DataSourceReference,
)
from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.elements.dimension import Dimension, DimensionType
from metricflow.model.objects.elements.identifier import Identifier, IdentifierType
from metricflow.model.objects.elements.measure import Measure
from metricflow.model.spec_converters import MeasureConverter
from metricflow.specs import (
    TimeDimensionSpec,
    DimensionSpec,
    IdentifierSpec,
    ColumnAssociationResolver,
    DEFAULT_TIME_GRANULARITY,
    IdentifierReference,
)
from metricflow.sql.sql_exprs import (
    SqlStringExpression,
    SqlDateTruncExpression,
    SqlColumnReferenceExpression,
    SqlColumnReference,
    SqlExpressionNode,
)
from metricflow.sql.sql_plan import (
    SqlTableFromClauseNode,
    SqlSelectStatementNode,
    SqlSelectColumn,
    SqlQueryPlanNode,
    SqlSelectQueryFromClauseNode,
)
from metricflow.time.time_granularity import TimeGranularity

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class DimensionConversionResult:
    """Helper class for returning the results of converting dimensions from a data source."""

    dimension_instances: Sequence[DimensionInstance]
    time_dimension_instances: Sequence[TimeDimensionInstance]
    select_columns: Sequence[SqlSelectColumn]


class DataSourceToDataSetConverter:
    """Converts a data source in the model to a data set that can be used with the dataflow plan builder.

    Identifier links generally refer to the identifiers used to join the measure source to the dimension source. For
    example, the dimension name "user_id__device_id__platform" has identifier links "user_id" and "device_id" and would
    mean that the measure source was joined by "user_id" to an intermediate source, and then it was joined by
    "device_id" to the source containing the "platform" dimension.
    """

    # Regex for inferring whether an expression for an element is a column reference.
    _SQL_IDENTIFIER_REGEX = re.compile("^[a-zA-Z_][a-zA-Z_0-9]*$")

    def __init__(self, column_association_resolver: ColumnAssociationResolver) -> None:  # noqa: D
        self._column_association_resolver = column_association_resolver

    def _create_dimension_instance(
        self,
        data_source_name: str,
        dimension: Dimension,
        identifier_links: Tuple[IdentifierReference, ...],
    ) -> DimensionInstance:
        """Create a dimension instance from the dimension object in the model."""
        dimension_spec = DimensionSpec(
            element_name=dimension.reference.element_name,
            identifier_links=identifier_links,
        )
        column_associations = dimension_spec.column_associations(self._column_association_resolver)

        return DimensionInstance(
            associated_columns=column_associations,
            spec=dimension_spec,
            defined_from=(
                DataSourceElementReference(
                    data_source_name=data_source_name,
                    element_name=dimension.reference.element_name,
                ),
            ),
        )

    def _create_time_dimension_instance(
        self,
        data_source_name: str,
        time_dimension: Dimension,
        identifier_links: Tuple[IdentifierReference, ...],
        time_granularity: TimeGranularity = DEFAULT_TIME_GRANULARITY,
    ) -> TimeDimensionInstance:
        """Create a time dimension instance from the dimension object from a data source in the model."""
        time_dimension_spec = TimeDimensionSpec(
            element_name=time_dimension.reference.element_name,
            identifier_links=identifier_links,
            time_granularity=time_granularity,
        )

        column_associations = time_dimension_spec.column_associations(self._column_association_resolver)

        return TimeDimensionInstance(
            associated_columns=column_associations,
            spec=time_dimension_spec,
            defined_from=(
                DataSourceElementReference(
                    data_source_name=data_source_name,
                    element_name=time_dimension.reference.element_name,
                ),
            ),
        )

    def _create_identifier_instance(
        self,
        data_source_name: str,
        identifier: Identifier,
        identifier_links: Tuple[IdentifierReference, ...],
    ) -> IdentifierInstance:
        """Create an identifier instance from the identifier object from a data sourcein the model."""
        identifier_spec = IdentifierSpec(
            element_name=identifier.reference.element_name,
            identifier_links=identifier_links,
        )
        column_associations = identifier_spec.column_associations(self._column_association_resolver)

        return IdentifierInstance(
            associated_columns=column_associations,
            spec=identifier_spec,
            defined_from=(
                DataSourceElementReference(
                    data_source_name=data_source_name,
                    element_name=identifier.reference.element_name,
                ),
            ),
        )

    @staticmethod
    def _make_element_sql_expr(
        table_alias: str, element_name: str, element_expr: Optional[str] = None
    ) -> SqlExpressionNode:
        """Create an expression that can be used for reading the element from the data source's SQL."""
        if element_expr:
            if DataSourceToDataSetConverter._SQL_IDENTIFIER_REGEX.match(element_expr) and element_expr.upper() not in (
                "TRUE",
                "FALSE",
                "NULL",
            ):
                return SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias=table_alias,
                        column_name=element_expr,
                    )
                )
            return SqlStringExpression(sql_expr=element_expr)

        return SqlColumnReferenceExpression(
            SqlColumnReference(
                table_alias=table_alias,
                column_name=element_name,
            )
        )

    def _convert_measures(
        self,
        data_source_name: str,
        measures: Sequence[Measure],
        measure_time_dimension_spec: TimeDimensionSpec,
        table_alias: str,
    ) -> Tuple[Sequence[MeasureInstance], Sequence[SqlSelectColumn]]:
        # Convert all elements to instances
        measure_instances = []
        select_columns = []
        for measure in measures or []:
            measure_spec = MeasureConverter.convert_to_measure_spec(measure=measure)
            measure_instance = MeasureInstance(
                associated_columns=measure_spec.column_associations(self._column_association_resolver),
                spec=measure_spec,
                defined_from=(
                    DataSourceElementReference(
                        data_source_name=data_source_name,
                        element_name=measure.reference.element_name,
                    ),
                ),
                aggregation_state=AggregationState.NON_AGGREGATED,
            )
            measure_instances.append(measure_instance)
            select_columns.append(
                SqlSelectColumn(
                    expr=DataSourceToDataSetConverter._make_element_sql_expr(
                        table_alias=table_alias,
                        element_name=measure.reference.element_name,
                        element_expr=measure.expr,
                    ),
                    column_alias=measure_instance.associated_column.column_name,
                )
            )

        return measure_instances, select_columns

    def _convert_dimensions(
        self,
        data_source_name: str,
        dimensions: Sequence[Dimension],
        identifier_links: Tuple[IdentifierReference, ...],
        table_alias: str,
    ) -> DimensionConversionResult:
        dimension_instances = []
        time_dimension_instances = []
        select_columns = []

        for dimension in dimensions or []:
            if dimension.type == DimensionType.CATEGORICAL:
                dimension_instance = self._create_dimension_instance(
                    data_source_name=data_source_name,
                    dimension=dimension,
                    identifier_links=identifier_links,
                )
                dimension_instances.append(dimension_instance)
                select_columns.append(
                    SqlSelectColumn(
                        expr=DataSourceToDataSetConverter._make_element_sql_expr(
                            table_alias=table_alias,
                            element_name=dimension.reference.element_name,
                            element_expr=dimension.expr,
                        ),
                        column_alias=dimension_instance.associated_column.column_name,
                    )
                )

            elif dimension.type == DimensionType.TIME:
                defined_time_granularity = TimeGranularity.DAY
                if dimension.type_params and dimension.type_params.time_granularity:
                    defined_time_granularity = dimension.type_params.time_granularity

                time_dimension_instance = self._create_time_dimension_instance(
                    data_source_name=data_source_name,
                    time_dimension=dimension,
                    identifier_links=identifier_links,
                    time_granularity=defined_time_granularity,
                )
                time_dimension_instances.append(time_dimension_instance)
                select_columns.append(
                    SqlSelectColumn(
                        expr=DataSourceToDataSetConverter._make_element_sql_expr(
                            table_alias=table_alias,
                            element_name=dimension.reference.element_name,
                            element_expr=dimension.expr,
                        ),
                        column_alias=time_dimension_instance.associated_column.column_name,
                    )
                )

                # Add time dimensions with a smaller granularity for ease in query resolution
                for time_granularity in TimeGranularity:
                    if time_granularity.to_int() > defined_time_granularity.to_int():
                        time_dimension_instance = self._create_time_dimension_instance(
                            data_source_name=data_source_name,
                            time_dimension=dimension,
                            identifier_links=identifier_links,
                            time_granularity=time_granularity,
                        )
                        time_dimension_instances.append(time_dimension_instance)

                        select_columns.append(
                            SqlSelectColumn(
                                expr=SqlDateTruncExpression(
                                    time_granularity=time_granularity,
                                    arg=DataSourceToDataSetConverter._make_element_sql_expr(
                                        table_alias=table_alias,
                                        element_name=dimension.reference.element_name,
                                        element_expr=dimension.expr,
                                    ),
                                ),
                                column_alias=time_dimension_instance.associated_column.column_name,
                            )
                        )
            else:
                assert False, f"Unhandled dimension type: {dimension.type}"

        return DimensionConversionResult(
            dimension_instances=dimension_instances,
            time_dimension_instances=time_dimension_instances,
            select_columns=select_columns,
        )

    def _create_identifier_instances(
        self,
        data_source_name: str,
        identifiers: Sequence[Identifier],
        identifier_links: Tuple[IdentifierReference, ...],
        table_alias: str,
    ) -> Tuple[Sequence[IdentifierInstance], Sequence[SqlSelectColumn]]:
        identifier_instances = []
        select_columns = []
        for identifier in identifiers or []:
            # We don't want to create something like user_id__user_id, so skip if the link is the same as the
            # identifier.
            if len(identifier_links) == 1 and identifier.reference == identifier_links[0]:
                continue

            identifier_instance = self._create_identifier_instance(
                data_source_name=data_source_name,
                identifier=identifier,
                identifier_links=identifier_links,
            )

            identifier_instances.append(identifier_instance)
            if identifier.is_composite:
                for idx in range(len(identifier.identifiers)):
                    sub_id = identifier.identifiers[idx]
                    column_name = identifier_instance.associated_columns[idx].column_name

                    expr = sub_id.expr
                    if expr is None:
                        assert sub_id.name is not None
                        expr = sub_id.name
                    sub_id_name = sub_id.ref or sub_id.name
                    assert sub_id_name, f"Sub-identifier {sub_id} must have 'name' or 'ref' defined"

                    select_columns.append(
                        SqlSelectColumn(
                            expr=DataSourceToDataSetConverter._make_element_sql_expr(
                                table_alias=table_alias,
                                element_name=sub_id_name,
                                element_expr=expr,
                            ),
                            column_alias=column_name,
                        )
                    )
            else:
                select_columns.append(
                    SqlSelectColumn(
                        expr=DataSourceToDataSetConverter._make_element_sql_expr(
                            table_alias=table_alias,
                            element_name=identifier.reference.element_name,
                            element_expr=identifier.expr,
                        ),
                        column_alias=identifier_instance.associated_column.column_name,
                    )
                )
        return identifier_instances, select_columns

    @staticmethod
    def _find_primary_time_dimension(data_source: DataSource) -> TimeDimensionSpec:
        # If a data source has measures, it should have a primary time dimension. Find it and use it to set the time
        # dimension for the measure.
        primary_time_dimensions = [dimension for dimension in data_source.dimensions if dimension.is_primary_time]
        assert len(primary_time_dimensions) == 1, (
            f"Data source ({data_source}) with measures should have exactly 1 primary time dimension, found "
            f"{len(primary_time_dimensions)}: {primary_time_dimensions}."
        )
        primary_time_dimension = primary_time_dimensions[0]
        assert (
            primary_time_dimension.type_params and primary_time_dimension.type_params.time_granularity
        ), f"Primary time dimension missing time granularity: {primary_time_dimension}"

        return TimeDimensionSpec(
            element_name=primary_time_dimension.reference.element_name,
            identifier_links=(),
            time_granularity=primary_time_dimension.type_params.time_granularity,
        )

    def create_sql_source_data_set(self, data_source: DataSource) -> DataSourceDataSet:
        """Create an SQL source data set from a data source in the model."""

        # Gather all instances and columns from all data sources.
        all_measure_instances: List[MeasureInstance] = []
        all_dimension_instances: List[DimensionInstance] = []
        all_time_dimension_instances: List[TimeDimensionInstance] = []
        all_identifier_instances: List[IdentifierInstance] = []

        all_select_columns: List[SqlSelectColumn] = []
        from_source_alias = IdGeneratorRegistry.for_class(self.__class__).create_id(f"{data_source.name}_src")

        # Handle measures
        if len(data_source.measures) > 0:
            primary_time_dimension = self._find_primary_time_dimension(data_source)
            measure_instances, select_columns = self._convert_measures(
                data_source_name=data_source.name,
                measures=data_source.measures,
                measure_time_dimension_spec=primary_time_dimension,
                table_alias=from_source_alias,
            )
            all_measure_instances.extend(measure_instances)
            all_select_columns.extend(select_columns)

        # For dimensions in a data source, we can access them through the local form, or the dundered form.
        # e.g. in the "users" data source, with the "country" dimension and the "user_id" identifier,
        # the dimensions "country" and "user_id__country" both mean the same thing. To make matching easier, create both
        # instances in the instance set. We'll create a different instance for each "possible_identifier_links".
        possible_identifier_links: List[Tuple[IdentifierReference, ...]] = [()]
        for identifier in data_source.identifiers:
            if identifier.type in (IdentifierType.PRIMARY, IdentifierType.UNIQUE):
                possible_identifier_links.append((identifier.reference,))

        # Handle dimensions
        conversion_results = [
            self._convert_dimensions(
                data_source_name=data_source.name,
                dimensions=data_source.dimensions,
                identifier_links=identifier_links,
                table_alias=from_source_alias,
            )
            for identifier_links in possible_identifier_links
        ]

        all_dimension_instances.extend(
            [
                dimension_instance
                for conversion_result in conversion_results
                for dimension_instance in conversion_result.dimension_instances
            ]
        )

        all_time_dimension_instances.extend(
            [
                time_dimension_instance
                for conversion_result in conversion_results
                for time_dimension_instance in conversion_result.time_dimension_instances
            ]
        )

        all_select_columns.extend(
            [
                select_column
                for conversion_result in conversion_results
                for select_column in conversion_result.select_columns
            ]
        )

        # Handle identifiers
        for identifier_links in possible_identifier_links:
            identifier_instances, select_columns = self._create_identifier_instances(
                data_source_name=data_source.name,
                identifiers=data_source.identifiers,
                identifier_links=identifier_links,
                table_alias=from_source_alias,
            )
            all_identifier_instances.extend(identifier_instances)
            all_select_columns.extend(select_columns)

        # Generate the "from" clause depending on whether it's an SQL query or an SQL table.
        from_source: Optional[SqlQueryPlanNode] = None
        if data_source.sql_table:
            from_source = SqlTableFromClauseNode(sql_table=SqlTable.from_string(data_source.sql_table))
        elif data_source.sql_query:
            from_source = SqlSelectQueryFromClauseNode(select_query=data_source.sql_query)
        else:
            raise RuntimeError(f"Data source does not have sql_table or sql_query set: {data_source}")

        select_statement_node = SqlSelectStatementNode(
            description=f"Read Elements From Data Source '{data_source.name}'",
            select_columns=tuple(all_select_columns),
            from_source=from_source,
            from_source_alias=from_source_alias,
            joins_descs=(),
            where=None,
            group_bys=(),
            order_bys=(),
        )

        return DataSourceDataSet(
            data_source_reference=DataSourceReference(data_source_name=data_source.name),
            instance_set=InstanceSet(
                measure_instances=tuple(all_measure_instances),
                dimension_instances=tuple(all_dimension_instances),
                time_dimension_instances=tuple(all_time_dimension_instances),
                identifier_instances=tuple(all_identifier_instances),
                metric_instances=(),
            ),
            sql_select_node=select_statement_node,
        )
